from __future__ import annotations

import os
import sys
from pathlib import Path
import weakref
from typing import TYPE_CHECKING
from contextlib import suppress

from qtpy.QtCore import Signal
from qtpy import QtWidgets as QtW, QtGui
from qtconsole.rich_jupyter_widget import RichJupyterWidget
from himena.utils.misc import lru_cache
from himena.qt._utils import get_stylesheet_path
from himena.plugins import validate_protocol

if TYPE_CHECKING:
    from himena.style import Theme
    from himena.widgets import MainWindow
    from himena_builtins.qt.console import ConsoleConfig

    class RichJupyterWidget(RichJupyterWidget, QtW.QWidget):
        """To fix typing problem"""

# Modified from napari_console https://github.com/napari/napari-console

if sys.platform.startswith("win"):
    import asyncio

    try:
        from asyncio import (
            WindowsProactorEventLoopPolicy,
            WindowsSelectorEventLoopPolicy,
        )
    except ImportError:
        pass
        # not affected
    else:
        if type(asyncio.get_event_loop_policy()) is WindowsProactorEventLoopPolicy:
            # WindowsProactorEventLoopPolicy is not compatible with tornado 6
            # fallback to the pre-3.8 default of Selector
            asyncio.set_event_loop_policy(WindowsSelectorEventLoopPolicy())


class QtConsole(RichJupyterWidget):
    """Console widget for the Python interpreter.

    The main window instance is passed to the console as variable `ui` by default.
    Tabs, windows, and the internal data models can be accessed from its attributes.

    ```python
    ui.tabs[0]  # get the first tab instance
    ui.tabs[0][1]  # get the second window in the first tab instance
    ui.current_window  # get the current sub-window instance
    ui.current_model  # get the current data model instance
    ```

    New data can be added to the GUI by its methods.

    ```python
    ui.add_tab()  # add a new tab
    ui.add_data_model(WidgetDataModel(...))  # add a new data model
    ui.add_object(..., type=...)  # add a new python object
    """

    codeExecuted = Signal()
    _instance = None

    def __init__(self, ui: MainWindow):
        super().__init__()
        self.setMinimumSize(100, 0)
        self.resize(100, 40)
        self._parent_connected = False
        self._main_window_symbol = "ui"
        self._matplotlib_backend = "inline"
        self._matplotlib_backend_orig = os.environ.get("MPLBACKEND")
        self._ui = ui
        self.codeExecuted.connect(self.setFocus)

    @classmethod
    def get_or_create(cls, ui) -> QtConsole:
        if cls._instance is None:
            cls._instance = cls(ui)
        return cls._instance

    def connect_parent(self):
        from IPython import get_ipython
        from IPython.terminal.interactiveshell import TerminalInteractiveShell
        from ipykernel.connect import get_connection_file
        from ipykernel.inprocess.ipkernel import InProcessInteractiveShell
        from ipykernel.zmqshell import ZMQInteractiveShell
        from qtconsole.client import QtKernelClient
        from qtconsole.inprocess import QtInProcessKernelManager

        if self._parent_connected:
            return
        self._parent_connected = True
        shell = get_ipython()

        if shell is None:
            # If there is no currently running instance create an in-process
            # kernel.
            kernel_manager = QtInProcessKernelManager()
            kernel_manager.start_kernel(show_banner=False)
            kernel_manager.kernel.gui = "qt"

            kernel_client = kernel_manager.client()
            kernel_client.start_channels()

            self.kernel_manager = kernel_manager
            self.kernel_client = kernel_client
            self.shell: InProcessInteractiveShell = kernel_manager.kernel.shell
            self.push = self.shell.push

        elif type(shell) == InProcessInteractiveShell:
            # If there is an existing running InProcessInteractiveShell
            # it is likely because multiple viewers have been launched from
            # the same process. In that case create a new kernel.
            # Connect existing kernel
            kernel_manager = QtInProcessKernelManager(kernel=shell.kernel)
            kernel_client = kernel_manager.client()

            self.kernel_manager = kernel_manager
            self.kernel_client = kernel_client
            self.shell = kernel_manager.kernel.shell
            self.push = self.shell.push

        elif isinstance(shell, TerminalInteractiveShell):
            # if launching from an ipython terminal then adding a console is
            # not supported. Instead users should use the ipython terminal for
            # the same functionality.
            self.kernel_client = None
            self.kernel_manager = None
            self.shell = None
            self.push = lambda var: None

        elif isinstance(shell, ZMQInteractiveShell):
            # if launching from jupyter notebook, connect to the existing
            # kernel
            kernel_client = QtKernelClient(connection_file=get_connection_file())
            kernel_client.load_connection_file()
            kernel_client.start_channels()

            self.kernel_manager = None
            self.kernel_client = kernel_client
            self.shell = shell
            self.push = self.shell.push
        else:
            raise ValueError(f"ipython shell not recognized, got {type(shell)}")

        if self.shell is not None:
            from IPython.paths import get_ipython_dir

            _exit = _get_exit_auto_call()
            _exit.set_main_window(self._ui)
            self.shell.push({"exit": _exit})  # update the "exit"

            # run IPython startup files
            profile_dir = Path(get_ipython_dir()) / "profile_default" / "startup"
            if profile_dir.exists():
                import runpy

                _globals = {}
                for startup in profile_dir.glob("*.py"):
                    with suppress(Exception):
                        _globals.update(runpy.run_path(str(startup)))

                self.shell.push(_globals)

            ns = {self._main_window_symbol: self._ui}
            self.shell.push(ns)

    def execute(self, source=None, hidden=False, interactive=False):
        out = super().execute(source, hidden, interactive)
        self.codeExecuted.emit()
        return out

    def setFocus(self):
        """Set focus to the text edit."""
        self._control.setFocus()
        return None

    def showEvent(self, event):
        """Show event."""
        super().showEvent(event)
        self.setFocus()
        os.environ["MPLBACKEND"] = self._matplotlib_backend
        return None

    def hideEvent(self, a0):
        if self._matplotlib_backend_orig is None:
            del os.environ["MPLBACKEND"]
        else:
            os.environ["MPLBACKEND"] = self._matplotlib_backend_orig
        return super().hideEvent(a0)

    @validate_protocol
    def widget_added_callback(self):
        self.connect_parent()
        QtW.QApplication.processEvents()

    @validate_protocol
    def theme_changed_callback(self, theme: Theme):
        """Update the console theme."""
        # need to set stylesheet via style_sheet property
        self.style_sheet = theme.format_text(get_stylesheet_path().read_text())

        # Set syntax styling and highlighting using theme
        if theme.is_light_background():
            self.syntax_style = "default"
        else:
            self.syntax_style = "native"
        bracket_color = QtGui.QColor(theme.highlight_dim)
        self._bracket_matcher.format.setBackground(bracket_color)

    def update_configs(
        self,
        cfg: ConsoleConfig,
    ):
        old_symbol = self._main_window_symbol
        self._main_window_symbol = cfg.main_window_symbol
        self._matplotlib_backend = cfg.matplotlib_backend
        if self._parent_connected:
            if (ui := self.shell.user_ns.get(old_symbol)) is self._ui:
                self.shell.drop_by_id({old_symbol: ui})
            self.shell.push({self._main_window_symbol: self._ui})


@lru_cache(maxsize=1)
def _get_exit_auto_call():
    from IPython.core.autocall import IPyAutocall

    class ExitAutocall(IPyAutocall):
        """Overwrite the default 'exit' autocall to close the viewer."""

        def __init__(self, ip=None):
            super().__init__(ip)
            self._main = None

        def set_main_window(self, window: MainWindow):
            self._main = weakref.ref(window)

        def __call__(self, *args, **kwargs):
            self._main().close()

    return ExitAutocall()



from __future__ import annotations
from typing import Callable, TypeVar

from magicgui.widgets import LineEdit, PushButton, Label
from magicgui.widgets.bases import ValuedContainerWidget

_V = TypeVar("_V")


def _label(text: str, width: int) -> Label:
    label = Label(value=text)
    label.max_width = width
    return label


class SliderRangeGetter(ValuedContainerWidget[tuple[_V, _V]]):
    """Widget to get a min and max value from an array widget."""

    def __init__(
        self,
        getter: Callable[[], _V],
        **kwargs,
    ):
        self._value_min = LineEdit()
        self._value_max = LineEdit()
        self._get_value_btn_min = PushButton(text="Read")
        self._get_value_btn_max = PushButton(text="Read")
        self._getter = getter
        self._value_min.max_width = 50
        self._value_max.max_width = 50
        self._get_value_btn_min.max_width = 40
        self._get_value_btn_max.max_width = 40

        super().__init__(
            widgets=[
                _label("min", 24),
                self._value_min,
                self._get_value_btn_min,
                _label("max", 24),
                self._value_max,
                self._get_value_btn_max,
            ],
            labels=False,
            layout="horizontal",
            **kwargs,
        )

        self._value_min.changed.connect(self._on_value_changed)
        self._get_value_btn_min.changed.connect(self._on_min_press)
        self._get_value_btn_max.changed.connect(self._on_max_press)
        self.margins = (0, 0, 0, 0)

    def _on_min_press(self):
        self._value_min.value = str(self._getter())

    def _on_max_press(self):
        self._value_max.value = str(self._getter())

    def _on_value_changed(self):
        self.changed.emit(self.get_value())

    def get_value(self) -> tuple[_V, _V]:
        if self._value_min.value.strip() == "":
            _value_min = None
        else:
            _value_min = int(self._value_min.value)
        if self._value_max.value.strip() == "":
            _value_max = None
        else:
            _value_max = int(self._value_max.value) + 1
        return (_value_min, _value_max)

    def set_value(self, value: tuple[_V, _V]):
        if value[0] is None:
            self._value_min.value = ""
        else:
            self._value_min.value = repr(value[0])
        if value[1] is None:
            self._value_max.value = ""
        else:
            # The max value is exclusive, so we need to subtract 1
            self._value_max.value = repr(value[1] - 1)



from himena.qt.main_window import MainWindowQt
from himena.qt.registry import register_widget_class
from himena.qt.magicgui import register_magicgui_types
from himena.qt import settings  # just register
from himena.qt._utils import (
    drag_model,
    drag_command,
    ndarray_to_qimage,
    qimage_to_ndarray,
)

__all__ = [
    "MainWindowQt",
    "register_widget_class",
    "drag_model",
    "drag_command",
    "ndarray_to_qimage",
    "qimage_to_ndarray",
]

register_magicgui_types()
del register_magicgui_types, settings



from __future__ import annotations

from contextlib import suppress
import sys
import socket
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Generic, TypeVar
from psygnal import Signal
from himena.exceptions import ExceptionHandler
from himena.widgets import current_instance
from himena.core import new_window
from himena._socket import InterProcessData

if TYPE_CHECKING:
    from IPython import InteractiveShell
    from qtpy.QtWidgets import QApplication
    from warnings import WarningMessage

_A = TypeVar("_A")  # the backend application type


class EventLoopHandler(ABC, Generic[_A]):
    errored = Signal(Exception)
    warned = Signal(object)
    socket_activated = Signal(bytes)
    _instances: dict[str, QtEventLoopHandler] = {}

    def __init__(self, name: str, host: str = "localhost", port: int = 49200):
        self._name = name
        self._instances[name] = self
        self._server_socket: socket.socket | None = None
        self._host: str = host
        self._port: int = port

    @classmethod
    def create(cls, name: str, host: str = "localhost", port: int = 49200):
        if name not in cls._instances:
            cls._instances[name] = QtEventLoopHandler(name, host, port)
        return cls._instances[name]

    @abstractmethod
    def get_app(self) -> _A:
        """Get Application instance."""

    @abstractmethod
    def run_app(self):
        """Start the event loop."""

    def close_socket(self):
        """Close the socket."""
        if self._server_socket:
            with suppress(OSError):
                self._server_socket.shutdown(socket.SHUT_RDWR)
            self._server_socket.close()
            self._server_socket = None


def gui_is_active(event_loop: str) -> bool:
    """True only if "%gui **" magic is called in ipython kernel."""
    shell = get_ipython_shell()
    return shell and shell.active_eventloop == event_loop


class QtEventLoopHandler(EventLoopHandler["QApplication"]):
    _APP: QApplication | None = None

    def get_app(self):
        """Get QApplication."""
        self.gui_qt()
        app = self.instance()
        if app is None:
            app = self.create_application()
        self._APP = app
        return app

    def create_application(self) -> QApplication:
        from qtpy.QtCore import Qt
        from qtpy.QtWidgets import QApplication
        from qtpy import QT6

        if not QT6:
            QApplication.setAttribute(Qt.ApplicationAttribute.AA_EnableHighDpiScaling)
        return QApplication([])

    def run_app(self):
        """Start the event loop."""
        if not gui_is_active("qt"):
            with ExceptionHandler(
                hook=self._except_hook,
                warning_hook=self._warn_hook,
            ) as _:
                qapp = self.get_app()
                self._setup_socket(qapp)
                qapp.exec()
                return None

        return self.get_app().exec()

    def _setup_socket(self, app):
        """Set up a socket for inter-process communication."""
        from qtpy.QtCore import QSocketNotifier

        self._server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self._server_socket.bind((self._host, self._port))
        self._server_socket.listen(1)
        self._qnotifier = QSocketNotifier(
            self._server_socket.fileno(),
            QSocketNotifier.Type.Read,
            parent=app,
        )
        self._qnotifier.activated.connect(self._on_socket_activated)

    def _on_socket_activated(self, socket: int):
        """Handle socket activation."""

        client_socket, _ = self._server_socket.accept()
        with client_socket:
            chunks = []
            while True:
                chunk = client_socket.recv(1024)
                if not chunk:
                    break
                chunks.append(chunk)
            incoming = b"".join(chunks)
        try:
            data = InterProcessData.from_bytes(incoming)
        except Exception as e:
            print("Failed to parse socket data: %s", e)
            return None
        try:
            ins = current_instance(data.profile_name)
        except KeyError:
            ins = new_window(data.profile_name)
        ins.show()
        for file in data.files:
            ins.read_file(file)

    def instance(self) -> QApplication | None:
        """Get QApplication instance or None if it does not exist."""
        from qtpy.QtWidgets import QApplication

        return QApplication.instance()

    def gui_qt(self) -> None:
        """Call "%gui qt" magic."""
        if not gui_is_active("qt"):
            shell = get_ipython_shell()
            if shell and shell.active_eventloop != "qt":
                shell.enable_gui("qt")
        return None

    def _except_hook(self, exc_type: type[Exception], exc_value: Exception, exc_tb):
        """Exception hook used during application execution."""
        return self.errored.emit(exc_value)

    def _warn_hook(self, warning: WarningMessage):
        """Warning hook used during application execution."""
        return self.warned.emit(warning)


class EmptyEventLoopHandler(EventLoopHandler):
    def get_app(self):
        return None

    def run_app(self):
        return None


def get_event_loop_handler(
    backend: str,
    app_name: str,
    host: str = "localhost",
    port: int = 49200,
) -> EventLoopHandler:
    if backend == "qt":
        return QtEventLoopHandler.create(app_name, host, port)
    else:
        return EmptyEventLoopHandler.create(app_name, host, port)


def get_ipython_shell() -> InteractiveShell | None:
    """Get ipython shell if available."""
    if "IPython" in sys.modules:
        from IPython import get_ipython

        return get_ipython()
    else:
        return None



from concurrent.futures import Future
from typing import Iterator, Literal, Any, cast, Union, TYPE_CHECKING
import uuid
from pydantic_compat import BaseModel, Field
from himena.workflow._base import WorkflowStep

if TYPE_CHECKING:
    from himena.types import WidgetDataModel
    from himena.workflow import Workflow


class CommandParameterBase(BaseModel):
    """A class that describes a parameter of a command."""

    type: str
    name: str
    value: Any

    def __repr_args__(self):
        for arg in super().__repr_args__():
            if arg[0] == "type":
                continue
            yield arg


class UserParameter(CommandParameterBase):
    """A class that describes a parameter that was set by a user."""

    type: Literal["user"] = "user"
    value: Any
    """Any python object for this parameter"""


class ModelParameter(CommandParameterBase):
    """A class that describes a parameter that was set by a model."""

    type: Literal["model"] = "model"
    value: uuid.UUID
    """workflow ID"""
    model_type: str


class WindowParameter(CommandParameterBase):
    """A class that describes a parameter that was set by a window."""

    type: Literal["window"] = "window"
    value: uuid.UUID
    """workflow ID"""
    model_type: str


class ListOfModelParameter(CommandParameterBase):
    """A class that describes a list of model parameters."""

    type: Literal["list"] = "list"
    value: list[uuid.UUID]
    """workflow IDs"""
    is_window: bool = False
    """True if the parameter is a list of windows."""


def parse_parameter(name: str, value: Any) -> "tuple[CommandParameterBase, Workflow]":
    """Normalize a k=v argument to a CommandParameterBase instance."""
    from himena.types import WidgetDataModel
    from himena.widgets import SubWindow
    from himena.workflow import Workflow

    if isinstance(value, WidgetDataModel):
        param = ModelParameter(
            name=name, value=value.workflow.last_id(), model_type=value.type
        )
        wf = value.workflow
    elif isinstance(value, SubWindow):
        model = value.to_model()
        wf = model.workflow
        param = WindowParameter(name=name, value=wf.last_id(), model_type=model.type)
    elif isinstance(value, list) and all(
        isinstance(each, WidgetDataModel) for each in value
    ):
        value_ = cast(list[WidgetDataModel], value)
        param = ListOfModelParameter(
            name=name, value=[each.workflow.last_id() for each in value_]
        )
        wf = Workflow.concat([each.workflow for each in value_])
    elif isinstance(value, list) and all(isinstance(each, SubWindow) for each in value):
        value_ = cast(list[SubWindow], value)
        models = [each.to_model() for each in value_]
        param = ListOfModelParameter(
            name=name,
            value=[each.workflow.last_id() for each in models],
            is_window=True,
        )
        wf = Workflow.concat([each.workflow for each in models])
    else:
        param = UserParameter(name=name, value=value)
        wf = Workflow()
    return param, wf


CommandParameterType = Union[
    UserParameter, ModelParameter, WindowParameter, ListOfModelParameter
]


class CommandExecution(WorkflowStep):
    """Describes that one was created by a command."""

    type: Literal["command"] = "command"
    command_id: str
    contexts: list[CommandParameterType] = Field(default_factory=list)
    parameters: list[CommandParameterType] | None = Field(
        default=None,
        description="Parameters passed to the command. None if the command is not parametric.",
    )
    execution_time: float = Field(default=0.0)  # seconds

    def iter_parents(self) -> Iterator[uuid.UUID]:
        for ctx in self.contexts:
            if isinstance(ctx, ModelParameter):
                yield ctx.value
            elif isinstance(ctx, WindowParameter):
                yield ctx.value

        for param in self.parameters or []:
            if isinstance(param, ModelParameter):
                # TODO: check if WindowParameter is needed here.
                yield param.value
            elif isinstance(param, ListOfModelParameter):
                yield from param.value

    def _get_model_impl(self, wf: "Workflow") -> "WidgetDataModel":
        from himena.types import WidgetDataModel
        from himena.widgets import current_instance

        model_context = None
        window_context = None
        ui = current_instance()
        for _ctx in self.contexts:
            if isinstance(_ctx, ModelParameter):
                model_context = wf.model_for_id(_ctx.value)
            elif isinstance(_ctx, WindowParameter):
                window_context = wf.window_for_id(_ctx.value)
                model_context = window_context.to_model()
            else:
                raise ValueError(
                    f"Context parameter must be a model: {_ctx} (command ID: {self.command_id})"
                )

        if self.parameters is None:
            params = None
        else:
            params = {}
            for _p in self.parameters:
                if isinstance(_p, UserParameter):
                    params[_p.name] = _p.value
                elif isinstance(_p, ModelParameter):
                    params[_p.name] = wf.filter(_p.value).model_for_id(_p.value)
                elif isinstance(_p, WindowParameter):
                    params[_p.name] = wf.filter(_p.value).window_for_id(_p.value)
                elif isinstance(_p, ListOfModelParameter):
                    if _p.is_window:
                        params[_p.name] = [
                            wf.filter(each).window_for_id(each) for each in _p.value
                        ]
                    else:
                        params[_p.name] = [
                            wf.filter(each).model_for_id(each) for each in _p.value
                        ]
                else:  # pragma: no cover
                    raise ValueError(
                        f"Unknown parameter type: {_p} (command ID: {self.command_id})"
                    )
        result = ui.exec_action(
            self.command_id,
            window_context=window_context,
            model_context=model_context,
            with_params=params,
            process_model_output=False,
        )
        if isinstance(result, Future):
            result = result.result()
        if not isinstance(result, WidgetDataModel):
            raise ValueError(
                f"Expected to return a WidgetDataModel but got {result} (command ID: {self.command_id})"
            )
        if main := wf._mock_main_window:
            win = main.add_data_model(result)
            win._identifier = self.id
        return result


class UserModification(WorkflowStep):
    """Describes that one was modified from another model."""

    type: Literal["user-modification"] = "user-modification"
    original: uuid.UUID

    def _get_model_impl(self, sf: "Workflow") -> "WidgetDataModel":
        # just skip modification...
        return sf.model_for_id(self.original)

    def iter_parents(self) -> Iterator[uuid.UUID]:
        yield self.original



from functools import partial
import inspect
from pathlib import Path
import html
import re
from typing import Any, Iterable, Literal, Mapping
import warnings
from datetime import datetime
import numpy as np
from himena.data_wrappers._dataframe import wrap_dataframe
from himena._descriptors import NoNeedToSave
from himena.plugins import (
    register_function,
    configure_gui,
    widget_classes,
    configure_submenu,
)
from himena.standards.model_meta import ListMeta
from himena.types import Parametric, WidgetDataModel
from himena.consts import StandardType, MonospaceFontFamily, MenuId
from himena.utils.collections import OrderedSet
from himena.widgets import SubWindow, MainWindow, ParametricWindow
from himena.workflow import Workflow, as_function
from himena import AppContext as ctx
from himena._utils import get_display_name, get_widget_class_id, unwrap_lazy_model

configure_submenu(MenuId.TOOLS_OTHERS, group="20_builtins", order=51)
configure_submenu(MenuId.TOOLS_FUNCTION, group="20_builtins", order=50)


@register_function(
    types=StandardType.READER_NOT_FOUND,
    menus=[],
    command_id="builtins:open-as-text-anyway",
)
def open_as_text_anyway(ui: MainWindow, win: SubWindow) -> WidgetDataModel[str]:
    """Open as a text file."""
    model = win.to_model()
    if model.type != StandardType.READER_NOT_FOUND:
        raise ValueError(f"Invalid model type: {model.type}")
    if not isinstance(src := model.source, Path):
        raise ValueError(
            f"Model has multiple or no local source paths: {src}. Cannot open as a text data."
        )
    out = model.with_value(
        src.read_text(),
        type=StandardType.TEXT,
        save_behavior_override=NoNeedToSave(),
    )
    out.extension_default = src.suffix
    win._close_me(ui)
    return out


@register_function(
    menus=[MenuId.TOOLS_OTHERS],
    title="Stack models ...",
    command_id="builtins:models:stack-models",
    enablement=(ctx.num_tabs > 0) & (ctx.num_sub_windows > 0),
)
def stack_models(ui: MainWindow) -> Parametric:
    """Stack windows into a single window that contains a list of the models."""

    def run_stack_models(
        models: list[WidgetDataModel],
        pattern: str = "",
    ) -> WidgetDataModel:
        """Stack windows into a single window that contains a list of the models.

        Parameters
        ----------
        models : list of WidgetDataModel
            Models to stack.
        pattern : str
            A regular expression pattern to match the title of the models.
        """
        if models and pattern:
            raise ValueError("Cannot specify both models and patterns.")
        if not models and not pattern:
            raise ValueError("Must specify either models or patterns.")
        if pattern:
            pat = re.compile(pattern)
            models: list[WidgetDataModel] = []
            for win in ui.tabs.current():
                if isinstance(win, ParametricWindow):
                    continue
                model = win.to_model()
                if pat.match(model.title):
                    models.append(model)
            if not models:
                raise ValueError("No models matched the pattern.")
        return WidgetDataModel(
            value=models,
            type=StandardType.MODELS,
            title="Merged",
        )

    return run_stack_models


@register_function(
    menus=[MenuId.TOOLS_OTHERS],
    types=[StandardType.MODELS],
    command_id="builtins:models:sort-model-list",
)
def sort_model_list(model: WidgetDataModel) -> Parametric:
    """Sort the model list."""

    def run_sort(
        descending: bool = False,
        sort_by: Literal["title", "type", "time"] = "title",
    ) -> WidgetDataModel:
        """Run sorting.

        Parameters
        ----------
        descending : bool
            Sort by descending order.
        sort_by : str
            Specify how to sort the list.
        """
        if sort_by == "title":
            _sort_func = _sort_func_title
        elif sort_by == "type":
            _sort_func = _sort_func_type
        elif sort_by == "time":
            _sort_func = _sort_func_datetime
        else:
            raise ValueError(f"Invalid `sort_by` argument: {sort_by}")
        models = sorted(
            _norm_model_list(model.value), key=_sort_func, reverse=descending
        )
        return model.with_value(models, title=f"{model.title} sorted")

    return run_sort


def _sort_func_title(m: WidgetDataModel) -> str:
    return m.title


def _sort_func_type(m: WidgetDataModel) -> str:
    return m.type


def _sort_func_datetime(m: WidgetDataModel) -> datetime:
    if last := m.workflow.last():
        return last.datetime
    return datetime(9999, 12, 31)


@register_function(
    menus=[MenuId.TOOLS_OTHERS],
    types=[StandardType.MODELS],
    command_id="builtins:models:filter-model-list",
)
def filter_model_list(model: WidgetDataModel) -> Parametric:
    """Filter the model list."""

    def run_filter(
        model_type: str = "",
        title_contains: str = "",
        unwrap_lazy_objects: bool = True,
    ) -> WidgetDataModel:
        """Filter the model list.

        Parameters
        ----------
        model_type : str
            If specified, only models of this type will be included.
        title_contains : str
            If specified, only models with titles containing this string will be
            included.
        unwrap_lazy_objects : bool
            If True, lazy-type models will be unwrapped before filtering. If you added
            a element from a local file, it is usually a lazy object.
        """
        models_out: list[WidgetDataModel] = []
        for m in _norm_model_list(model.value):
            if not isinstance(m, WidgetDataModel):
                warnings.warn(
                    f"Expected a sequence of WidgetDataModel but got {type(m)} as an "
                    "element. Skipping.",
                    RuntimeWarning,
                    stacklevel=2,
                )
                continue
            if unwrap_lazy_objects and m.type == StandardType.LAZY:
                m = unwrap_lazy_model(m)
            if model_type and not m.is_subtype_of(model_type):
                continue
            if title_contains and title_contains not in m.title:
                continue
            models_out.append(m)
        return model.with_value(models_out, title=f"{model.title} filtered")

    return run_filter


@register_function(
    title="Compute lazy items",
    types=[StandardType.MODELS],
    menus=[MenuId.TOOLS_OTHERS],
    command_id="builtins:models:compute-lazy-items",
)
def compute_lazy_items(model: WidgetDataModel) -> WidgetDataModel:
    """Compute all the lazy items in the model list."""

    def _unwrap_list(mlist: list[WidgetDataModel]):
        out: list[WidgetDataModel] = []
        for m in _norm_model_list(mlist):
            if m.type == StandardType.LAZY:
                out.append(unwrap_lazy_model(m))
            elif m.is_subtype_of(StandardType.MODELS):
                out.append(m.with_value(_unwrap_list(m.value)))
            else:
                out.append(m)
        return out

    model_out = _unwrap_list(model.value)
    return model.with_value(model_out, title=f"{model.title} computed")


def _norm_model_list(models) -> Iterable[WidgetDataModel]:
    if isinstance(models, Mapping):
        return models.values()
    else:
        return models


@register_function(
    title="Execute workflow",
    types=[StandardType.WORKFLOW],
    menus=[],
    command_id="builtins:exec-workflow",
    enablement=ctx.active_window_model_subtype_1 != "parametric",
)
def exec_workflow(model: WidgetDataModel) -> None:
    """Execute the workflow."""
    # NOTE: this function should not return the model and let the application process
    # it. The workflow of the output model should not be updated. Using workflow to
    # compute a new model is a special case in himena.
    _cast_workflow(model).compute(process_output=True)


@register_function(
    title="Execute workflow",
    types=[StandardType.WORKFLOW_PARAMETRIC],
    menus=[],
    command_id="builtins:exec-parametric-workflow",
)
def exec_parametric_workflow(model: WidgetDataModel, ui: MainWindow) -> Parametric:
    """Execute the workflow."""
    wf = _cast_workflow(model)
    return as_function(wf)(ui)


def _cast_workflow(model: WidgetDataModel) -> Workflow:
    """Cast the model to a Workflow object."""
    if not isinstance(model.value, Workflow):
        raise TypeError(f"Expected a Workflow object but got {type(model.value)}")
    return model.value


@register_function(
    types=[
        StandardType.TEXT,
        StandardType.TABLE,
        StandardType.DATAFRAME,
        StandardType.ARRAY,
        StandardType.EXCEL,
    ],
    menus=["tools"],
    command_id="builtins:general:show-statistics",
)
def show_statistics(model: WidgetDataModel) -> WidgetDataModel:
    """Show the statistics of the data."""
    value = model.value
    if model.is_subtype_of(StandardType.TEXT):
        value_str = str(value)
        nchars = len(value_str)
        nlines = len(value_str.splitlines())
        out = (
            f"<b>Number of characters:</b> {nchars}<br><b>Number of lines:</b> {nlines}"
        )
    elif model.is_subtype_of(StandardType.TABLE):
        out = _statistics_table(value)
    elif model.is_subtype_of(StandardType.DATAFRAME):
        df = wrap_dataframe(value)
        nr, nc = df.shape
        columns = df.column_names()
        shape = f"<b>Shape:</b> {nr} rows, {nc} columns"
        dtypes = "".join(
            ["<li>"]
            + [f"<ul>{c!r}: {d.name}</ul>" for c, d in zip(columns, df.dtypes)]
            + ["</li>"]
        )
        stats = []
        for c in columns:
            ar = df.column_to_array(c)
            if ar.dtype.kind in "iuf":
                stats.append(
                    f"<b>{c!r}</b>: min={ar.min()}, max={ar.max()}, mean={ar.mean()}, std={ar.std(ddof=1)}"
                )
            elif ar.dtype.kind == "b":
                ntrue = ar.sum()
                stats.append(f"<b>{c!r}</b>: True ... {ntrue}/{len(ar)}")
        stats = "<br>" + "<br>".join(stats)
        out = shape + dtypes + stats
    elif model.is_subtype_of(StandardType.ARRAY):
        if not isinstance(value, np.ndarray):
            raise ValueError(f"Expected a numpy array but got {type(value)}")
        if value.dtype.names:
            out = f"<b>Shape:</b> {value.shape}"
        else:
            out = f"<b>Shape:</b> {value.shape}<br><b>Min:</b> {value.min()}<br><b>Max:</b> {value.max()}<br><b>Mean:</b> {value.mean()}<br><b>Std:</b> {value.std(ddof=1)}"
    elif model.is_subtype_of(StandardType.EXCEL):
        value = model.value
        if not isinstance(value, dict):
            raise ValueError(f"Expected a dict but got {type(out)}")
        out = []
        for key, val in value.items():
            out.append(f"<h3><u>{key}</u></h3>{_statistics_table(val)}")
        out = "".join(out)
    else:
        raise NotImplementedError(f"Statistics for {model.type} is not implemented.")
    return WidgetDataModel(
        value=out,
        type=StandardType.HTML,
        title=f"Statistics of {model.title}",
        extension_default=".html",
        editable=False,
        save_behavior_override=NoNeedToSave(),
    )


@register_function(
    menus=["tools"],
    command_id="builtins:general:show-metadata",
)
def show_metadata(model: WidgetDataModel) -> WidgetDataModel:
    """Show the metadata of the underlying data."""
    meta = model.metadata
    if meta is None:
        out = "<No metadata>"
    elif hasattr(meta, "_repr_html_"):
        out = meta._repr_html_()
    else:
        meta_repr = html.escape(repr(meta))
        out = (
            f"<span style='font-family: monaco,{MonospaceFontFamily},"
            f"monospace;'>{meta_repr}</span>"
        )
    return WidgetDataModel(
        value=out,
        type=StandardType.HTML,
        title=f"Metadata of {model.title}",
        editable=False,
        save_behavior_override=NoNeedToSave(),
    )


@register_function(
    title="Specify widget ...",
    menus=["tools"],
    command_id="builtins:specify-widget",
)
def specify_widget(model: WidgetDataModel) -> Parametric:
    """Manually specify the type of the data.

    This function will force the data to be open with the specified widget. Whether the
    widget supports the value will not be checked.
    """
    classes = sorted(widget_classes().items(), key=lambda x: x[0])
    choices = [
        (f"{get_display_name(c, sep=' ')}", c)
        for t, c in classes
        if t != StandardType.READER_NOT_FOUND
    ]

    @configure_gui(widget_class={"choices": choices}, show_parameter_labels=False)
    def run_specify(widget_class: type) -> WidgetDataModel:
        return model.with_open_plugin(
            open_with=get_widget_class_id(widget_class),
            save_behavior_override=NoNeedToSave(),
        )

    return run_specify


@register_function(
    title="Partialize function ...",
    menus=[MenuId.TOOLS_FUNCTION],
    types=[StandardType.FUNCTION],
    command_id="builtins:partialize-function",
)
def partialize_function(model: WidgetDataModel) -> Parametric:
    """Partialize the function."""
    import ast

    if not callable(func := model.value):
        raise ValueError(f"Expected a callable object but got {type(func)}")

    options = {}
    sig = inspect.signature(func)
    for name, param in sig.parameters.items():
        if param.default is param.empty:
            options[name] = {"label": name, "widget_type": "LineEdit", "value": ""}
        else:
            options[name] = {
                "label": name,
                "widget_type": "LineEdit",
                "value": str(param.default),
            }

    @configure_gui(gui_options=options)
    def run_partialize(**kwargs: str) -> WidgetDataModel:
        kwargs_evaled = {k: ast.literal_eval(v) for k, v in kwargs.items() if v.strip()}
        return WidgetDataModel(
            value=partial(func, **kwargs_evaled),
            type=StandardType.FUNCTION,
            title=f"[Partial] {model.title}",
        )

    return run_partialize


@register_function(
    title="Plot y = f(x)...",
    menus=[MenuId.TOOLS_FUNCTION],
    types=[StandardType.FUNCTION],
    command_id="builtins:plot-function-1d",
)
def plot_function_1d(model: WidgetDataModel) -> Parametric:
    """Plot function by its first argument."""

    def run_plot(
        xmin: float = -1, xmax: float = 1, num_points: int = 100
    ) -> WidgetDataModel:
        func = model.value
        x = np.linspace(xmin, xmax, num_points)
        y = func(x)
        return WidgetDataModel(
            value={"x": x, "y": y},
            type=StandardType.DATAFRAME_PLOT,
            title=f"Plot of {model.title}",
        )

    return run_plot


@register_function(
    title="Plot z = f(x, y)...",
    menus=[MenuId.TOOLS_FUNCTION],
    types=[StandardType.FUNCTION],
    command_id="builtins:plot-function-2d",
)
def plot_function_2d(model: WidgetDataModel) -> Parametric:
    """Plot function by its first two arguments."""
    from himena.standards import plotting as hplt

    def run_plot(
        xmin: float = -1,
        xmax: float = 1,
        ymin: float = -1,
        ymax: float = 1,
        num_points: int = 100,
    ) -> WidgetDataModel:
        func = model.value
        xvals = np.linspace(xmin, xmax, num_points)
        yvals = np.linspace(ymin, ymax, num_points)
        x, y = np.meshgrid(xvals, yvals)
        z = func(x, y)
        fig = hplt.figure_3d()
        fig.axes.surface(x, y, z)
        return WidgetDataModel(
            value=fig,
            type=StandardType.PLOT,
            title=f"Plot of {model.title}",
        )

    return run_plot


@register_function(
    title="Results To Table",
    menus=[MenuId.TOOLS_OTHERS],
    types=[StandardType.RESULTS],
    command_id="builtins:results:results-to-table",
)
def results_to_table(model: WidgetDataModel) -> WidgetDataModel:
    """Convert the result stack to a table."""
    vals: list[dict[str, Any]] = model.value
    return _results_to_table(vals, title=f"{model.title} (Table)")


@register_function(
    title="Results To Table (Selected)",
    menus=[MenuId.TOOLS_OTHERS],
    types=[StandardType.RESULTS],
    command_id="builtins:results:selected-results-to-table",
)
def selected_results_to_table(model: WidgetDataModel) -> WidgetDataModel:
    """Convert the selected results to a table."""
    meta = model.metadata
    if not isinstance(meta, ListMeta):
        raise ValueError(f"Expected a ListMeta but got {type(meta)}")
    vals: list[dict[str, Any]] = [model.value[i] for i in meta.selections]
    if not vals:
        raise ValueError("No item selected.")
    return _results_to_table(vals, title=f"{model.title} (Table)")


def _results_to_table(vals: list[dict[str, Any]], title: str) -> WidgetDataModel:
    """Convert the result stack to a table."""
    columns = OrderedSet[str]()
    for val in vals:
        if not isinstance(val, dict):
            raise TypeError(f"Expected a dict but got {type(val)}")
        columns.update(val.keys())
    column_name_to_index = {name: i for i, name in enumerate(columns)}
    arr = np.zeros((len(vals) + 1, len(columns)), dtype=np.dtypes.StringDType())
    arr[0, :] = [str(name) for name in columns]
    for i, val in enumerate(vals):
        for name, value in val.items():
            if name not in column_name_to_index:
                raise ValueError(f"Unknown column name: {name}")
            arr[i + 1, column_name_to_index[name]] = str(value)
    return WidgetDataModel(
        value=arr,
        type=StandardType.TABLE,
        title=title,
    )


def _statistics_table(value) -> str:
    if value is None:
        value = np.array([[""]], dtype=np.dtypes.StringDType())
    if not isinstance(value, np.ndarray):
        raise ValueError(
            f"Expected a numpy array for the table data but got {type(value)}"
        )
    if not isinstance(value.dtype, np.dtypes.StringDType):
        raise ValueError(f"Expected a numpy array of strings but got {value.dtype}")
    nrows, ncols = value.shape
    n_empty = int((value == "").ravel().sum())
    nchars = len(value.ravel().sum())
    out = (
        f"<b>Number of rows:</b> {nrows}<br>"
        f"<b>Number of columns:</b> {ncols}<br>"
        f"<b>Number of empty cells:</b> {n_empty}<br>"
        f"<b>Number of non-empty cells:</b> {nrows * ncols - n_empty}<br>"
        f"<b>Number of characters:</b> {nchars}"
    )
    return out



from __future__ import annotations

from qtpy import QtWidgets as QtW, QtGui
from himena.types import WidgetDataModel
from himena.consts import MonospaceFontFamily
from himena.plugins import validate_protocol


class QFallbackWidget(QtW.QPlainTextEdit):
    """A fallback widget for the data of non-registered type."""

    def __init__(self):
        super().__init__()
        self.setReadOnly(True)
        self.setFont(QtGui.QFont(MonospaceFontFamily))
        self._model: WidgetDataModel | None = None

    @validate_protocol
    def update_model(self, model: WidgetDataModel):
        self.setPlainText(f"type: {model.type!r}\nvalue:\n{model.value!r}")
        self._model = model
        return

    @validate_protocol
    def to_model(self) -> WidgetDataModel:
        if self._model is None:
            raise ValueError("Model is not set")
        return self._model

    @validate_protocol
    def model_type(self) -> str:
        return self._model.type



from __future__ import annotations

from qtpy import QtWidgets as QtW, QtCore
from qtpy.QtCore import Qt, Signal
from himena.types import DockArea, DockAreaString


class QDockWidget(QtW.QDockWidget):
    closed = QtCore.Signal()
    whats_this = QtCore.Signal()

    def __init__(
        self,
        widget: QtW.QWidget,
        title: str,
        allowed_areas: list[DockAreaString | DockArea] | None = None,
    ):
        super().__init__(title)
        self.setWidget(widget)
        _titlebar = QDockWidgetTitleBar(title, self)
        self.setTitleBarWidget(_titlebar)
        _titlebar.closeSignal.connect(self.close)
        _titlebar.whatsThisSignal.connect(self.whats_this.emit)
        if allowed_areas is None:
            allowed_areas = [
                DockArea.LEFT,
                DockArea.RIGHT,
                DockArea.TOP,
                DockArea.BOTTOM,
            ]
        else:
            allowed_areas = [DockArea(area) for area in allowed_areas]
        areas = QtCore.Qt.DockWidgetArea.NoDockWidgetArea
        for allowed_area in allowed_areas:
            areas |= _DOCK_AREA_MAP[allowed_area]
        return self.setAllowedAreas(areas)

    @staticmethod
    def area_normed(area) -> QtCore.Qt.DockWidgetArea:
        if area is not None:
            area = DockArea(area)
        return _DOCK_AREA_MAP[area]

    def closeEvent(self, event):
        self.closed.emit()
        super().closeEvent(event)


_DOCK_AREA_MAP = {
    DockArea.TOP: QtCore.Qt.DockWidgetArea.TopDockWidgetArea,
    DockArea.BOTTOM: QtCore.Qt.DockWidgetArea.BottomDockWidgetArea,
    DockArea.LEFT: QtCore.Qt.DockWidgetArea.LeftDockWidgetArea,
    DockArea.RIGHT: QtCore.Qt.DockWidgetArea.RightDockWidgetArea,
    None: QtCore.Qt.DockWidgetArea.NoDockWidgetArea,
}


class QDockWidgetTitleFrame(QtW.QFrame):
    """The title bar horizontal line."""

    def __init__(self):
        super().__init__()
        self.setFrameShape(QtW.QFrame.Shape.HLine)
        self.setFrameShadow(QtW.QFrame.Shadow.Sunken)
        self.setSizePolicy(
            QtW.QSizePolicy.Policy.Expanding, QtW.QSizePolicy.Policy.Fixed
        )


class QDockWidgetTitleBar(QtW.QWidget):
    """A custom title bar for a dock widget"""

    whatsThisSignal = Signal()
    closeSignal = Signal()

    def __init__(self, title: str = "", parent: QtW.QWidget | None = None) -> None:
        super().__init__(parent)
        _layout = QtW.QHBoxLayout(self)
        _layout.setContentsMargins(4, 0, 4, 0)
        _layout.setSpacing(0)

        self._title_label = QtW.QLabel()
        self._title_label.setContentsMargins(0, 0, 0, 0)

        _frame = QDockWidgetTitleFrame()
        self._close_button = QtW.QToolButton()
        self._close_button.setText("✕")
        self._close_button.setToolTip("Close the widget.")
        self._close_button.setFixedSize(QtCore.QSize(16, 16))
        self._close_button.setCursor(Qt.CursorShape.ArrowCursor)

        self._whats_this_button = QtW.QToolButton()
        self._whats_this_button.setText("?")
        self._whats_this_button.setToolTip("What's this widget?")
        self._whats_this_button.setFixedSize(QtCore.QSize(16, 16))
        self._whats_this_button.setCursor(Qt.CursorShape.ArrowCursor)

        _layout.addWidget(self._title_label)
        _layout.addWidget(_frame)
        _layout.addWidget(self._whats_this_button)
        _layout.addWidget(self._close_button)
        _layout.setAlignment(self._close_button, Qt.AlignmentFlag.AlignRight)

        self._whats_this_button.clicked.connect(self.whatsThisSignal.emit)
        self._close_button.clicked.connect(self.closeSignal.emit)

        self.setTitle(title)
        self.setFixedHeight(16)
        self.setCursor(Qt.CursorShape.OpenHandCursor)

    def title(self) -> str:
        """The title text."""
        return self._title_label.text()

    def setTitle(self, text: str):
        """Set the title text."""
        if text == "":
            self._title_label.setVisible(False)
        else:
            self._title_label.setVisible(True)
            self._title_label.setText(f"  {text}  ")

    def setBold(self, bold: bool):
        """Set the title text bold."""
        font = self._title_label.font()
        font.setBold(bold)
        self._title_label.setFont(font)



from __future__ import annotations
from pathlib import Path

from qtpy import QtWidgets as QtW, QtCore

from himena.consts import StandardType
from himena.types import WidgetDataModel
from himena.plugins import validate_protocol
from himena.qt._utils import get_main_window


class QReaderNotFound(QtW.QWidget):
    """No reader function is defined for the file.

    The content of the file is not read yet. This widget only contains the file path.
    """

    __himena_widget_id__ = "builtins:QReaderNotFound"

    def __init__(self):
        super().__init__()
        layout = QtW.QVBoxLayout(self)
        self._label = QtW.QLabel("Reader not found.")
        self._label.setTextInteractionFlags(
            QtCore.Qt.TextInteractionFlag.TextSelectableByMouse
        )
        self._label.setWordWrap(True)
        self._open_as_text_button = QtW.QPushButton("Open as text")
        self._open_as_text_button.clicked.connect(self._open_as_text)
        self._file_path: Path | None = None

        layout.addWidget(self._label)
        layout.addWidget(self._open_as_text_button)

    @validate_protocol
    def update_model(self, model: WidgetDataModel[Path]):
        self._file_path = model.value
        if not self._file_path.exists():
            self._label.setText(f"File not found: {self._file_path}")
            return
        _byte = self._file_path.stat().st_size
        if _byte < 1024:
            _size = f"{_byte} B"
        elif _byte < 1024**2:
            _size = f"{_byte / 1024:.2f} KB"
        elif _byte < 1024**3:
            _size = f"{_byte / 1024 ** 2:.2f} MB"
        else:
            _size = f"{_byte / 1024 ** 3:.2f} GB"
        self._label.setText(f"Reader not found for {model.value.name} ({_size})")

    @validate_protocol
    def to_model(self) -> WidgetDataModel[Path]:
        return WidgetDataModel(
            value=self._file_path,
            type=self.model_type(),
        )

    @validate_protocol
    def model_type(self) -> str:
        return StandardType.READER_NOT_FOUND

    def _open_as_text(self):
        get_main_window(self).exec_action("builtins:open-as-text-anyway")



from typing import Literal, Union
from pydantic_compat import BaseModel, Field, field_validator
import numpy as np
from numpy.typing import NDArray

from himena.utils.misc import ansi2html


class IpynbStreamOutput(BaseModel):
    output_type: Literal["stream"] = "stream"
    name: str = Field(default="stdout")
    text: list[str] = Field(default_factory=list)

    def get_text_plain(self) -> str:
        return "".join(self.text)


class IpynbDisplayDataOutput(BaseModel):
    output_type: Literal["display_data"] = "display_data"
    data: dict = Field(default_factory=dict)
    metadata: dict = Field(default_factory=dict)

    def get_text_plain(self) -> str:
        return "\n".join(self.data.get("text/plain", []))

    def get_text_html(self) -> str:
        return "<br>".join(self.data.get("text/html", []))

    def get_image(self) -> NDArray[np.uint8] | None:
        import base64
        from PIL import Image
        import io

        if "image/png" in self.data:
            img_text = self.data["image/png"]
        elif "image/jpeg" in self.data:
            img_text = self.data["image/jpeg"]
        else:
            return None
        png_data = base64.b64decode(img_text, validate=True)
        buf = io.BytesIO(png_data)
        with Image.open(buf) as img:
            arr = np.array(img.convert("RGBA")).copy()
        return arr


class IpynbExecuteResultOutput(IpynbDisplayDataOutput):
    output_type: Literal["execute_result"] = "execute_result"


class IpynbErrorOutput(BaseModel):
    output_type: Literal["error"] = "error"
    ename: str = Field(default="")
    evalue: str = Field(default="")
    traceback: list[str] = Field(default_factory=list)

    def get_html(self, is_dark: bool = False) -> str:
        _html = (
            "\n".join(self.traceback)
            .replace(" ", "&nbsp;")
            .replace("<", "&lt;")
            .replace(">", "&gt;")
            .replace("\n", "<br>")
        )
        out = "".join(ansi2html(_html, is_dark=is_dark))
        return out


OutputTypes = Union[
    IpynbStreamOutput,
    IpynbDisplayDataOutput,
    IpynbExecuteResultOutput,
    IpynbErrorOutput,
]


class IpynbCell(BaseModel):
    """Model for a single cell in a Jupyter Notebook."""

    cell_type: str = Field(
        default="code",
        description="Type of cell, such as 'code' or 'markdown'",
    )
    metadata: dict = Field(default_factory=dict)
    source: str = Field("")
    outputs: list[OutputTypes] = Field(default_factory=list)

    @field_validator("source", mode="before")
    def _source_to_str(cls, v) -> str:
        if isinstance(v, list):
            return "".join(v)
        return v


class IpynbMetadata(BaseModel):
    kernel_info: dict = Field(default_factory=dict)
    language_info: dict = Field(default_factory=dict)


class IpynbFile(BaseModel):
    metadata: IpynbMetadata = Field(default_factory=IpynbMetadata)
    nbformat: int | None = Field(None)
    nbformat_minor: int | None = Field(None)
    cells: list[IpynbCell] = Field(default_factory=list)

    @property
    def language(self) -> str:
        lang = self.metadata.language_info.get("name")
        if lang is None:
            lang = "python"
        return lang



from __future__ import annotations

from typing import TYPE_CHECKING
from qtpy import QtWidgets as QtW, QtCore, QtGui

from himena.types import WindowState

if TYPE_CHECKING:
    from himena.qt.main_window import QMainWindow


class QWindowListWidget(QtW.QListWidget):
    def __init__(self):
        super().__init__()
        self.setSelectionBehavior(QtW.QAbstractItemView.SelectionBehavior.SelectRows)
        self.setCursor(QtCore.Qt.CursorShape.PointingHandCursor)


def line_edit(text: str):
    line = QtW.QLineEdit()
    line.setText(text)
    line.setEnabled(False)
    font = line.font()
    font.setPointSize(12)
    font.setBold(True)
    line.setFont(font)
    line.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
    return line


BIG_INT = 999999


class QGotoWidget(QtW.QWidget):
    def __init__(self, main: QMainWindow):
        super().__init__(main)
        self._main = main
        self._list_widgets: list[QWindowListWidget] = []
        self._stack = QtW.QStackedWidget()
        layout = QtW.QVBoxLayout(self)
        layout.addWidget(self._stack)
        self.close()

    def update_ui(self):
        while self._stack.count() > 0:
            self._stack.removeWidget(self._stack.widget(0))
        self._list_widgets.clear()
        main = self._main._himena_main_window
        tab = main.tabs.current()
        if tab is None:
            raise ValueError("No tab is opened.")
        for i_tab, tab in main.tabs.enumerate():
            area = QtW.QWidget()
            layout = QtW.QVBoxLayout(area)
            layout.addWidget(line_edit(f"({i_tab}) {tab.name}"))
            list_widget = QWindowListWidget()
            for i_win, win in tab.enumerate():
                list_widget.addItem(win.title)
            layout.addWidget(list_widget)
            self._stack.addWidget(area)
            list_widget.itemClicked.connect(self.activate_window_for_item)
            self._list_widgets.append(list_widget)
        lw = self.currentListWidget()
        if (idx := main.tabs.current_index) is not None:
            self._stack.setCurrentIndex(idx)
        if (idx := main.tabs.current().current_index) is not None:
            lw.setCurrentRow(idx)
        lw.setFocus()

    def currentListWidget(self) -> QtW.QListWidget:
        return self._list_widgets[self._stack.currentIndex()]

    def show(self) -> None:
        cur = self._main._tab_widget.currentWidget()
        self.update_ui()
        center = self._main.mapFromGlobal(cur.mapToGlobal(cur.rect().center()))
        dx = min(cur.width(), 280)
        dy = min(cur.height(), 270)
        rect = QtCore.QRect(center.x() - dx // 2, center.y() - dy // 2, dx, dy)
        self.setGeometry(rect)
        super().show()
        self._force_list_item_selected()
        return None

    def keyPressEvent(self, a0: QtGui.QKeyEvent | None) -> None:
        if a0 is None:
            return
        nr = self.currentListWidget().count()
        nc = self._stack.count()
        _ctrl = a0.modifiers() & QtCore.Qt.KeyboardModifier.ControlModifier
        move = BIG_INT if _ctrl else 1
        if a0.key() == QtCore.Qt.Key.Key_Escape:
            self.close()
        elif a0.key() == QtCore.Qt.Key.Key_Return:
            self.activate_window_for_current_index()
        elif a0.key() == QtCore.Qt.Key.Key_Up:
            self.currentListWidget().setCurrentRow(
                max(self.currentListWidget().currentRow() - move, 0)
            )
            self._force_list_item_selected()
            self.setFocus()
        elif a0.key() == QtCore.Qt.Key.Key_Down:
            self.currentListWidget().setCurrentRow(
                min(self.currentListWidget().currentRow() + move, nr - 1)
            )
            self._force_list_item_selected()
            self.setFocus()
        elif a0.key() == QtCore.Qt.Key.Key_Left:
            self._stack.setCurrentIndex(max(self._stack.currentIndex() - move, 0))
            self._force_list_item_selected()
            self.setFocus()
        elif a0.key() == QtCore.Qt.Key.Key_Right:
            self._stack.setCurrentIndex(min(self._stack.currentIndex() + move, nc - 1))
            self._force_list_item_selected()
            self.setFocus()
        else:
            return super().keyPressEvent(a0)

    def _force_list_item_selected(self):
        lw = self.currentListWidget()
        lw.setCurrentRow(max(lw.currentRow(), 0))

    def activate_window_for_item(self, item: QtW.QListWidgetItem | None = None):
        if item is None:
            self.close()
            return
        return self.activate_window_for_current_index()

    def activate_window_for_current_index(self):
        i_tab = self._stack.currentIndex()
        main = self._main._himena_main_window
        main.tabs.current_index = i_tab
        if self.currentListWidget().count() > 0:
            main.tabs.current().current_index = self.currentListWidget().currentRow()
        if (win := main.current_window) and win.state is WindowState.MIN:
            win.state = WindowState.NORMAL
        self.close()
        return None

    def focusOutEvent(self, a0: QtGui.QFocusEvent) -> None:
        self.close()
        return super().focusOutEvent(a0)



from dataclasses import dataclass, field


@dataclass
class MatplotlibCanvasConfigs:
    """Matplotlib canvas configurations."""

    font_size: int = field(default=10)
    font_family: str = field(default="Arial")

    def to_dict(self) -> dict:
        return {
            "font.size": self.font_size,
            "font.family": self.font_family,
        }



from __future__ import annotations

import numpy as np
from numpy.typing import NDArray
from typing import Callable, Iterator, TYPE_CHECKING, Generic, TypeVar
from dataclasses import dataclass, field

if TYPE_CHECKING:
    from typing import Self

_T = TypeVar("_T")
_U = TypeVar("_U")


def _item_factory():
    return np.empty((0,), dtype=np.object_)


def _indices_factory():
    return np.empty((0, 0), dtype=np.int32)


@dataclass
class NDObjectCollection(Generic[_T]):
    """List of nd objects, with useful methods."""

    items: NDArray[np.object_] = field(default_factory=_item_factory)
    indices: NDArray[np.int32] = field(default_factory=_indices_factory)
    axis_names: list[str] = field(default_factory=list)

    def __post_init__(self):
        if not isinstance(self.items, np.ndarray):
            self.items = np.asarray(self.items, dtype=np.object_)
        if self.items.ndim != 1:
            raise ValueError(
                f"Items must be a 1D array, got {self.items} ({self.items.ndim}D)."
            )
        if self.indices.shape[1] == 0:
            self.indices = np.empty((len(self.items), 0), dtype=np.int32)
        elif self.indices.shape[0] != len(self.items):
            raise ValueError("Indices must have the same length as items.")
        if self.indices.shape[0] == 0:
            self.indices = np.empty((0, len(self.axis_names)), dtype=np.int32)

    @property
    def ndim(self) -> int:
        """Number of dimensions of the collection."""
        return self.indices.shape[1]

    def set_axis_names(self, names: list[str]):
        """Update the axis names and coerce array shapes."""
        if self.indices.shape[0] == 0:
            self.indices = np.empty((0, len(names)), dtype=np.uint32)
        else:
            if self.ndim != len(names):
                raise ValueError(f"Expected {self.ndim} axis names, got {len(names)}")
        self.axis_names = names

    def coerce_dimensions(self, target_axis_names: list[str]) -> Self:
        """Reformat the collection to match the target axis names."""
        if self.axis_names == target_axis_names:
            return self
        columns = []
        for aname in target_axis_names:
            if aname in self.axis_names:
                index = self.axis_names.index(aname)
                columns.append(self.indices[:, index])
            else:
                columns.append(-np.ones(len(self.items), dtype=np.int32))
        indices = np.column_stack(columns)
        return self.__class__(
            items=self.items,
            indices=indices,
            axis_names=target_axis_names,
        )

    def mask_by_indices(self, key: tuple[int, ...]) -> NDArray[np.bool_] | None:
        """Binary mask for the given indices (None if nothing to mask out)."""
        if self.indices is None:
            raise ValueError("Indices are not set.")
        if len(self.axis_names) != len(key):
            raise ValueError(f"Expected {self.ndim} indices, got {len(key)}")
        ok = None
        for column_index, value in enumerate(key):
            column = self.indices[:, column_index]
            contains = np.logical_or(column == value, column < 0)
            if ok is None:
                ok = contains
            else:
                ok = ok & contains
        return ok

    def filter_by_indices(self, key: tuple[int, ...]) -> Self:
        ok = self.mask_by_indices(key)
        if ok is None:
            return self.__class__(
                items=self.items,
                indices=self.indices,
                axis_names=self.axis_names,
            )
        else:
            return self.__class__(
                items=self.items[ok],
                indices=self.indices[ok],
                axis_names=self.axis_names,
            )

    def filter_by_selection(self, selection: NDArray[np.bool_] | list[int]) -> Self:
        return self.__class__(
            items=self.items[selection],
            indices=self.indices[selection],
            axis_names=self.axis_names,
        )

    def add_item(self, indices, item: _T) -> None:
        """Add item at the given indice"""
        self.items = np.append(self.items, item)
        indices = np.atleast_2d(indices)
        self.indices = np.append(self.indices, indices, axis=0)

    def extend(self, other: Self) -> None:
        if len(self) > 0:
            if self.axis_names != other.axis_names:
                raise ValueError("Axis names must match.")
        if len(self) == 0:
            self.items = other.items.copy()
            self.indices = other.indices.copy()
        else:
            self.items = np.concatenate([self.items, other.items], axis=0)
            self.indices = np.concatenate([self.indices, other.indices], axis=0)
        return None

    def pop(self, index: int) -> _T:
        item = self.items[index]
        self.items = np.delete(self.items, index, axis=0)
        self.indices = np.delete(self.indices, index, axis=0)
        return item

    def clear(self) -> None:
        self.items = _item_factory()
        self.indices = _indices_factory()

    def copy(self) -> Self:
        return self.__class__(
            items=self.items.copy(),
            indices=self.indices.copy(),
            axis_names=self.axis_names.copy(),
        )

    def __getitem__(self, key: int) -> _T:
        return self.items[key]

    def __iter__(self) -> Iterator[_T]:
        return iter(self.items)

    def iter_with_indices(self) -> Iterator[tuple[tuple[int, ...], _T]]:
        for indices, item in zip(self.indices, self.items):
            yield tuple(indices), item

    def __len__(self) -> int:
        return len(self.items)

    def take_axis(self, axis: int, index: int):
        if axis >= len(self.axis_names):
            return self  # this happens when image is RGB image
        axis_name = self.axis_names[axis]
        column = self.indices[:, axis]
        ok = np.logical_or(column == index, column < 0)
        items = self.items[ok]
        indices = self.indices[ok]
        axis_names = [name for name in self.axis_names if name != axis_name]
        return self.__class__(
            items=items,
            indices=indices,
            axis_names=axis_names,
        )

    def project(self, axis: int):
        axis_name = self.axis_names[axis]
        items = self.items
        indices = self.indices
        axis_names = [name for name in self.axis_names if name != axis_name]
        return self.__class__(
            items=items,
            indices=indices,
            axis_names=axis_names,
        )

    def map_elements(
        self,
        func: Callable[[_T], _U],
        into: type[NDObjectCollection] | None = None,
    ) -> NDObjectCollection[_U]:
        if into is None:
            into = self.__class__
        return into(
            items=[func(item) for item in self.items],
            indices=self.indices,
            axis_names=self.axis_names,
        )

    def simplified(self) -> Self:
        """Drop axis"""
        cannot_drop = np.all(self.indices >= 0, axis=0)
        indices = np.take(self.indices, cannot_drop, axis=1)
        axis_names = [a for i, a in enumerate(self.axis_names) if i in cannot_drop]
        return self.__class__(
            items=self.items,
            indices=indices,
            axis_names=axis_names,
        )



from himena.data_wrappers._dataframe import wrap_dataframe
from himena.standards.model_meta import TableMeta
from himena.plugins import register_function, configure_gui, configure_submenu
from himena.types import WidgetDataModel, Parametric
from himena.consts import StandardType, MenuId
from himena.core import create_model

configure_submenu(MenuId.TOOLS_DATAFRAME, group="20_builtins", order=2)


@register_function(
    types=StandardType.DATAFRAME,
    menus=[MenuId.TOOLS_DATAFRAME],
    command_id="builtins:dataframe:header-to-row",
)
def header_to_row(model: WidgetDataModel) -> WidgetDataModel:
    """Convert the header as the first row."""
    df = wrap_dataframe(model.value)
    if isinstance(meta := model.metadata, TableMeta):
        sep = meta.separator or ","
    else:
        sep = ","
    csv_string = df.to_csv_string(sep)
    new_columns = [f"column_{i}" for i in range(df.num_columns())]
    input_string = f"{sep.join(new_columns)}\n{csv_string}"
    new_value = df.from_csv_string(input_string, sep)
    return model.with_value(new_value, update_inplace=True)


@register_function(
    title="Series as array",
    types=StandardType.DATAFRAME,
    menus=[MenuId.TOOLS_DATAFRAME],
    command_id="builtins:dataframe:series-as-array",
)
def series_as_array(model: WidgetDataModel) -> Parametric:
    """Convert a single column to an array."""

    def _get_column_index(*_):
        c0 = _get_column_selection_name(model)
        if c0 is None:
            raise ValueError("Please select a single column.")
        return c0

    @configure_gui(column={"bind": _get_column_index})
    def run_series_as_array(column: str):
        return create_model(
            wrap_dataframe(model.value).column_to_array(column),
            title=f"{model.title} ({column})",
            type=StandardType.ARRAY,
            extension_default=".npy",
        )

    return run_series_as_array


@register_function(
    title="Select columns by name ...",
    types=StandardType.DATAFRAME,
    menus=[MenuId.TOOLS_DATAFRAME],
    command_id="builtins:dataframe:select-columns-by-name",
)
def select_columns_by_name(model: WidgetDataModel) -> Parametric:
    """Select a subset of columns by name and return a new DataFrame."""
    df = wrap_dataframe(model.value)

    @configure_gui(
        columns={"choices": df.column_names(), "widget_type": "Select"},
    )
    def run(columns: list[str]):
        df_new = wrap_dataframe(model.value).select(columns).unwrap()
        return model.with_value(df_new).with_title_numbering()

    return run


@register_function(
    title="Filter DataFrame ...",
    types=StandardType.DATAFRAME,
    menus=[MenuId.TOOLS_DATAFRAME],
    command_id="builtins:dataframe:filter",
    run_async=True,
)
def filter_dataframe(model: WidgetDataModel) -> Parametric:
    import operator as _op

    choices = [
        ("Equal (==)", "eq"), ("Not Equal (!=)", "ne"), ("Greater (>)", "gt"),
        ("Greater Equal (>=)", "ge"), ("Less (<)", "lt"), ("Less Equal (<=)", "le"),
    ]  # fmt: skip
    df = wrap_dataframe(model.value)
    column_names = df.column_names()
    selected_column_name = _get_column_selection_name(model)

    column_name_option = {"choices": column_names}
    if selected_column_name is not None:
        column_name_option["value"] = selected_column_name

    @configure_gui(
        column=column_name_option,
        operator={"choices": choices},
    )
    def run(column: str, operator: str, value: str):
        op_func = getattr(_op, operator)
        series = df[column]
        if series.dtype.kind in "iuf":
            value_parsed = float(value)
        elif series.dtype.kind == "b":
            value_parsed = value.lower() in ["true", "1"]
        elif series.dtype.kind == "c":
            value_parsed = complex(value)
        else:
            value_parsed = value
        sl = op_func(series, value_parsed)
        df_new = df.filter(sl).unwrap()
        return model.with_value(df_new).with_title_numbering()

    return run


@register_function(
    title="Sort DataFrame ...",
    types=StandardType.DATAFRAME,
    menus=[MenuId.TOOLS_DATAFRAME],
    command_id="builtins:dataframe:sort",
    run_async=True,
)
def sort_dataframe(model: WidgetDataModel) -> Parametric:
    """Sort the DataFrame by a column."""
    df = wrap_dataframe(model.value)
    column_names = df.column_names()

    @configure_gui(column={"choices": column_names})
    def run(column: str, descending: bool = False, inplace: bool = False):
        df_new = df.sort(column, descending=descending).unwrap()
        return model.with_value(df_new, update_inplace=inplace).with_title_numbering()

    return run


@register_function(
    title="New column using function ...",
    types=StandardType.DATAFRAME,
    menus=[MenuId.TOOLS_DATAFRAME],
    command_id="builtins:dataframe:new-column-using-function",
)
def new_column_using_function(model: WidgetDataModel) -> Parametric:
    """Add a new column using a user-defined function."""
    df = wrap_dataframe(model.value)

    @configure_gui(
        input_column_name={"choices": df.column_names()},
        function={"types": [StandardType.FUNCTION]},
    )
    def run(
        input_column_name: str,
        function: WidgetDataModel,
        output_column_name: str,
        inplace: bool = False,
    ):
        """Run function and add the result as a new column.

        Parameters
        ----------
        input_column_name : str
            The name of the input column.
        function : WidgetDataModel
            The function to apply to the input column.
        output_column_name : str
            The name of the output column.
        inplace : bool
            If True, update the widget in place, otherwise create a new widget.
        """
        df = wrap_dataframe(model.value)
        col = df[input_column_name]
        out = function.value(col)
        dict_ = df.to_dict()
        dict_[output_column_name] = out
        df_new = df.from_dict(dict_)
        return model.with_value(df_new, update_inplace=inplace).with_title_numbering()

    return run


def _get_column_selection_name(model: WidgetDataModel) -> str | None:
    if isinstance(meta := model.metadata, TableMeta):
        if len(meta.selections) == 1:
            df = wrap_dataframe(model.value)
            (r0, r1), (c0, c1) = meta.selections[0]
            if r0 == 0 and r1 == df.num_rows() and c1 - c0 == 1:
                return df.column_names()[c0]
    return None



from himena.qt.magicgui import SelectionEdit, SliderRangeGetter

def test_string_parsing():
    widget = SelectionEdit(value=((3, 6), (5, 10)))
    assert widget.value == ((3, 6), (5, 10))
    assert widget._line_edit.value == "3:6, 5:10"
    widget.value = (None, 3), (6, None)
    assert widget.value == ((None, 3), (6, None))
    assert widget._line_edit.value == ":3, 6:"
    widget.value = None
    assert widget.value is None

def test_slider_range_getter():
    def _getter() -> int:
        return _getter.value
    _getter.value = 1
    widget = SliderRangeGetter(getter=_getter)
    assert widget.get_value() == (None, None)
    widget._get_value_btn_min.clicked()
    assert widget.get_value() == (1, None)
    _getter.value = 4
    widget._get_value_btn_max.clicked()
    assert widget.get_value() == (1, 5)
    assert widget._value_min.value == "1"
    assert widget._value_max.value == "4"  # inclusive max
    widget.set_value((4, 8))
    assert widget.get_value() == (4, 8)
    assert widget._value_min.value == "4"
    assert widget._value_max.value == "7"



from __future__ import annotations

import math
import logging
from typing import TYPE_CHECKING, Generic, TypeVar
from qtpy import QtGui, QtCore
from qtpy.QtCore import Qt
from himena_builtins.qt.widgets._image_components._roi_items import (
    QPointRoi,
    QPointsRoi,
    QPolygonRoi,
    QRoi,
    QLineRoi,
    QCircleRoi,
    QRectangleRoi,
    QEllipseRoi,
    QRotatedRectangleRoi,
    QRotatedEllipseRoi,
    QSegmentedLineRoi,
)
from himena_builtins.qt.widgets._image_components._handles import RoiSelectionHandles
from himena.standards.mouse import MouseEventHandler
from himena.widgets import set_status_tip

if TYPE_CHECKING:
    from ._graphics_view import QImageGraphicsView

_LOGGER = logging.getLogger(__name__)
_R = TypeVar("_R", bound=QRoi)
_R1 = TypeVar("_R1", bound=QSegmentedLineRoi)


class QtMouseEvent(MouseEventHandler[QtGui.QMouseEvent]):
    def __init__(self, view: QImageGraphicsView):
        self._view = view
        self._pos_drag_start: QtCore.QPoint | None = None
        self._pos_drag_prev: QtCore.QPoint | None = None

    @property
    def selection_handles(self) -> RoiSelectionHandles:
        return self._view._selection_handles

    def _mouse_move_pan_zoom(self, pos: QtCore.QPointF):
        delta = pos - self._pos_drag_prev
        self._view.move_items_by(delta.x(), delta.y())

    def pressed(self, event):
        self._pos_drag_start = event.pos()
        self._pos_drag_prev = event.pos()

    def released(self, event):
        self._pos_drag_prev = None
        self._pos_drag_start = None

    def is_click(self, event: QtGui.QMouseEvent) -> bool:
        """Check if the event is a click."""
        return (
            self._pos_drag_start is not None
            and (self._pos_drag_start - event.pos()).manhattanLength() < 2
        )


class PanZoomMouseEvents(QtMouseEvent):
    def pressed(self, event):
        super().pressed(event)
        self._view.viewport().setCursor(Qt.CursorShape.ClosedHandCursor)
        self._view.scene().setGrabSource(self)

    def moved(self, event):
        if self._pos_drag_start is None or self._pos_drag_prev is None:
            return
        self._mouse_move_pan_zoom(event.pos())
        self._pos_drag_prev = event.pos()

    def released(self, event):
        self._view.viewport().setCursor(Qt.CursorShape.ArrowCursor)
        return super().released(event)

    def double_clicked(self, event):
        self._view.fitInView(self._view.sceneRect(), Qt.AspectRatioMode.KeepAspectRatio)


class SelectMouseEvents(QtMouseEvent):
    def pressed(self, event):
        super().pressed(event)
        self._view.select_item_at(self._view.mapToScene(event.pos()))
        self._view.scene().setGrabSource(self)

    def moved(self, event):
        if self._pos_drag_prev is None:
            return
        pos = self._view.mapToScene(event.pos())
        if item := self._view._current_roi_item:
            if self._pos_drag_prev is None:
                return
            x0, y0 = self._view._pos_to_tuple(
                self._view.mapToScene(self._pos_drag_prev)
            )
            x1, y1 = self._view._pos_to_tuple(pos)
            delta = QtCore.QPointF(x1 - x0, y1 - y0)
            item.translate(delta.x(), delta.y())
            self._view.current_roi_updated.emit()
        else:
            # just forward to the pan-zoom mode
            self._mouse_move_pan_zoom(event.pos())
        self._pos_drag_prev = event.pos()


class RoiMouseEvents(QtMouseEvent, Generic[_R]):
    def roi(self) -> _R:
        out = self._view._current_roi_item
        assert isinstance(out, self.roi_type())
        return out

    def roi_type(self) -> type[_R]:
        raise NotImplementedError

    def make_roi(self, x: float, y: float, pen: QtGui.QPen) -> _R:
        raise NotImplementedError


class _SingleDragRoiMouseEvents(RoiMouseEvents[_R]):
    def _update_roi(self, start: QtCore.QPointF, current: QtCore.QPointF):
        raise NotImplementedError

    def _coerce_pos(
        self,
        pos: QtCore.QPointF,
        pos0: QtCore.QPointF,
    ) -> tuple[QtCore.QPointF, QtCore.QPointF]:
        return pos, pos0

    def pressed(self, event: QtGui.QMouseEvent):
        super().pressed(event)
        self._view.remove_current_item(reason="start drawing new ROI")
        p = self._view.mapToScene(self._pos_drag_start)
        self.make_roi(p.x(), p.y(), self._view._roi_pen)
        self.selection_handles.update_handle_size(self._view.transform().m11())

    def moved(self, event: QtGui.QMouseEvent):
        if self._pos_drag_start is None or self._pos_drag_prev is None:
            return
        pos = self._view.mapToScene(event.pos())
        if isinstance(self._view._current_roi_item, self.roi_type()):
            pos0 = self._view.mapToScene(self._pos_drag_start)
            if event.modifiers() & Qt.KeyboardModifier.ShiftModifier:
                pos, pos0 = self._coerce_pos(pos, pos0)
            self._update_roi(pos0, pos)
            self._view.current_roi_updated.emit()

    def released(self, event: QtGui.QMouseEvent):
        if event.button() == Qt.MouseButton.LeftButton:
            if self.is_click(event):
                self._view.remove_current_item(
                    reason=f"stop drawing {self._view.mode()}"
                )
            else:
                if cur_item := self._view._current_roi_item:
                    xscale = yscale = self._view._scale_bar_widget._scale
                    unit = self._view._scale_bar_widget._unit
                    set_status_tip(cur_item.short_description(xscale, yscale, unit))
        return super().released(event)

    def double_clicked(self, event):
        self._view.remove_current_item(reason="double-click")


class _LineTypeRoiMouseEvents(_SingleDragRoiMouseEvents[_R1]):
    def _coerce_pos(
        self,
        pos: QtCore.QPointF,
        pos0: QtCore.QPointF,
    ) -> tuple[QtCore.QPointF, QtCore.QPointF]:
        return _find_nice_position(pos, pos0), pos0


class _RectangleTypeRoiMouseEvents(_SingleDragRoiMouseEvents[_R]):
    def _coerce_pos(
        self,
        pos: QtCore.QPointF,
        pos0: QtCore.QPointF,
    ) -> tuple[QtCore.QPointF, QtCore.QPointF]:
        return _find_nice_rect_position(pos, pos0), pos0


class _MultiRoiMouseEvents(RoiMouseEvents[_R]):
    def pressed(self, event: QtGui.QMouseEvent):
        if (
            type(self._view._current_roi_item) is self.roi_type()
            and not self.selection_handles.is_drawing_polygon()
        ):
            # drawing the same ROI type, remove the previous one and end.
            self._view.remove_current_item(reason="start drawing new polygon")
            self._view.select_item(None)
        elif type(self._view._current_roi_item) is not self.roi_type():
            # just after switched from another ROI type, remove the previous one
            # and start drawing the current one.
            self._view.remove_current_item(reason="start drawing new ROI")
            self.selection_handles.start_drawing_polygon()
            self.selection_handles._is_last_vertex_added = True
        return super().pressed(event)

    def released(self, event: QtGui.QMouseEvent):
        if self._pos_drag_start is None:
            return
        if self._pos_drag_start != event.pos():
            return
        if self.selection_handles.is_drawing_polygon():
            p0 = self._view.mapToScene(self._pos_drag_start)
            item = self._view._current_roi_item
            if not isinstance(item, self.roi_type()):
                x1, y1 = p0.x(), p0.y()
                self.make_roi(x1, y1, self._view._roi_pen)
                self.selection_handles.update_handle_size(self._view.transform().m11())
            else:
                _LOGGER.info(
                    f"Added point {self._pos_drag_start} to {self._view.mode()}"
                )
                item.add_point(p0)
            self.selection_handles._is_last_vertex_added = True
            self._view.current_roi_updated.emit()
        return super().released(event)

    def double_clicked(self, event):
        self.selection_handles.finish_drawing_polygon()
        self._pos_drag_start = self._pos_drag_prev = None


class _SegmentedTypeRoiMouseEvents(_MultiRoiMouseEvents[_R1]):
    def moved(self, event: QtGui.QMouseEvent):
        pos = self._view.mapToScene(event.pos())
        if event.button() == Qt.MouseButton.NoButton:
            if self.selection_handles.is_drawing_polygon() and isinstance(
                item := self._view._current_roi_item, self.roi_type()
            ):
                # update the last point of the polygon
                if self.selection_handles._is_last_vertex_added:
                    self.selection_handles._is_last_vertex_added = False
                    item.add_point(pos)
                    self._view.current_roi_updated.emit()
                else:
                    num = item.count()
                    if num > 1:
                        item.update_point(num - 1, pos)
                        self._view.current_roi_updated.emit()


class PointsRoiMouseEvents(_MultiRoiMouseEvents[QPointsRoi]):
    def roi_type(self) -> type[QPointsRoi]:
        return QPointsRoi

    def make_roi(self, x: float, y: float, pen: QtGui.QPen) -> QPointsRoi:
        roi = QPointsRoi([x], [y]).withPen(pen)
        self._view.set_current_roi(roi)
        self.selection_handles.connect_points(roi)
        return roi

    def moved(self, event: QtGui.QMouseEvent):
        if event.button() == Qt.MouseButton.LeftButton:
            self._mouse_move_pan_zoom(event.pos())
            self._pos_drag_prev = event.pos()
            return
        return super().moved(event)


class SegmentedLineRoiMouseEvents(_SegmentedTypeRoiMouseEvents[QSegmentedLineRoi]):
    def roi_type(self) -> type[QSegmentedLineRoi]:
        return QSegmentedLineRoi

    def make_roi(self, x: float, y: float, pen: QtGui.QPen) -> QSegmentedLineRoi:
        roi = QSegmentedLineRoi([x], [y]).withPen(pen)
        self._view.set_current_roi(roi)
        self.selection_handles.connect_path(roi)
        return roi


class PolygonRoiMouseEvents(_SegmentedTypeRoiMouseEvents[QPolygonRoi]):
    def roi_type(self) -> type[QPolygonRoi]:
        return QPolygonRoi

    def make_roi(self, x: float, y: float, pen: QtGui.QPen) -> QPolygonRoi:
        roi = QPolygonRoi([x], [y]).withPen(pen)
        self._view.set_current_roi(roi)
        self.selection_handles.connect_path(roi)
        return roi


class LineRoiMouseEvents(_LineTypeRoiMouseEvents[QLineRoi]):
    def roi_type(self) -> type[QLineRoi]:
        return QLineRoi

    def make_roi(self, x: float, y: float, pen: QtGui.QPen) -> QLineRoi:
        roi = QLineRoi(x, y, x, y).withPen(pen)
        self._view.set_current_roi(roi)
        self.selection_handles.connect_line(roi)
        return roi

    def _update_roi(self, start, current):
        if isinstance(item := self._view._current_roi_item, QLineRoi):
            item.setLine(start.x(), start.y(), current.x(), current.y())


class RotatedRectangleRoiMouseEvents(_LineTypeRoiMouseEvents[QRotatedRectangleRoi]):
    def roi_type(self) -> type[QRotatedRectangleRoi]:
        return QRotatedRectangleRoi

    def make_roi(self, x: float, y: float, pen: QtGui.QPen) -> QRotatedRectangleRoi:
        p0 = QtCore.QPointF(x, y)
        roi = QRotatedRectangleRoi(p0, p0).withPen(pen)
        self._view.set_current_roi(roi)
        self.selection_handles.connect_rotated_rect(roi)
        return roi

    def _update_roi(self, start, current):
        if isinstance(item := self._view._current_roi_item, QRotatedRectangleRoi):
            item.set_end(current)


class RotatedEllipseRoiMouseEvents(_LineTypeRoiMouseEvents[QRotatedEllipseRoi]):
    def roi_type(self) -> type[QRotatedEllipseRoi]:
        return QRotatedEllipseRoi

    def make_roi(self, x: float, y: float, pen: QtGui.QPen) -> QRotatedEllipseRoi:
        p0 = QtCore.QPointF(x, y)
        roi = QRotatedEllipseRoi(p0, p0).withPen(pen)
        self._view.set_current_roi(roi)
        self.selection_handles.connect_rotated_rect(roi)
        return roi

    def _update_roi(self, start, current):
        if isinstance(item := self._view._current_roi_item, QRotatedEllipseRoi):
            item.set_end(current)


class RectangleRoiMouseEvents(_RectangleTypeRoiMouseEvents[QRectangleRoi]):
    def roi_type(self) -> type[QRectangleRoi]:
        return QRectangleRoi

    def make_roi(self, x: float, y: float, pen: QtGui.QPen) -> QRectangleRoi:
        roi = QRectangleRoi(x, y, 0, 0).withPen(pen)
        self._view.set_current_roi(roi)
        self.selection_handles.connect_rect(roi)
        return roi

    def _update_roi(self, start, current):
        if isinstance(item := self._view._current_roi_item, QRectangleRoi):
            x0, y0 = self._view._pos_to_tuple(current)
            x1, y1 = self._view._pos_to_tuple(start)
            width = abs(x1 - x0)
            height = abs(y1 - y0)
            item.setRect(min(x0, x1), min(y0, y1), width, height)


class CircleRoiMouseEvents(_SingleDragRoiMouseEvents[QCircleRoi]):
    def roi_type(self) -> type[QCircleRoi]:
        return QCircleRoi

    def make_roi(self, x: float, y: float, pen: QtGui.QPen) -> QCircleRoi:
        roi = QCircleRoi(x, y, 0).withPen(pen)
        self._view.set_current_roi(roi)
        self.selection_handles.connect_circle(roi)
        return roi

    def _update_roi(self, start: QtCore.QPointF, current: QtCore.QPointF):
        if isinstance(item := self._view._current_roi_item, QCircleRoi):
            x0, y0 = self._view._pos_to_tuple(start)
            x1, y1 = self._view._pos_to_tuple(current)
            radius = min(abs(x1 - x0), abs(y1 - y0))
            item.setCenterAndRadius((x0, y0), radius)


class EllipseRoiMouseEvents(_RectangleTypeRoiMouseEvents[QEllipseRoi]):
    def roi_type(self) -> type[QEllipseRoi]:
        return QEllipseRoi

    def make_roi(self, x: float, y: float, pen: QtGui.QPen) -> QEllipseRoi:
        roi = QEllipseRoi(x, y, 0, 0).withPen(pen)
        self._view.set_current_roi(roi)
        self.selection_handles.connect_rect(roi)
        return roi

    def _update_roi(self, start, current):
        if isinstance(item := self._view._current_roi_item, QEllipseRoi):
            x0, y0 = self._view._pos_to_tuple(current)
            x1, y1 = self._view._pos_to_tuple(start)
            width = abs(x1 - x0)
            height = abs(y1 - y0)
            item.setRect(min(x0, x1), min(y0, y1), width, height)


class PointRoiMouseEvents(RoiMouseEvents[QPointRoi]):
    def roi_type(self) -> type[QPointRoi]:
        return QPointRoi

    def make_roi(self, x: float, y: float, pen: QtGui.QPen) -> QPointRoi:
        roi = QPointRoi(x, y).withPen(pen)
        self._view.set_current_roi(roi)
        self.selection_handles.connect_point(roi)
        return roi

    def moved(self, event: QtGui.QMouseEvent):
        if event.button() == Qt.MouseButton.LeftButton:
            self._mouse_move_pan_zoom(event.pos())
            self._pos_drag_prev = event.pos()

    def released(self, event: QtGui.QMouseEvent):
        if (
            event.button() == Qt.MouseButton.LeftButton
            and self._pos_drag_start == event.pos()  # clicked
        ):
            pos = self._view.mapToScene(event.pos())
            self._view.remove_current_item(reason=f"stop drawing {self._view.mode()}")
            self.make_roi(pos.x(), pos.y(), self._view._roi_pen)
            self.selection_handles.update_handle_size(self._view.transform().m11())
        return super().released(event)


def _find_nice_position(pos: QtCore.QPointF, pos0: QtCore.QPointF) -> QtCore.QPointF:
    """Find the "nice" position when Shift is pressed."""
    x0, y0 = pos0.x(), pos0.y()
    x1, y1 = pos.x(), pos.y()
    ang = math.atan2(y1 - y0, x1 - x0)
    pi = math.pi
    if -pi / 8 < ang <= pi / 8:  # right direction
        y1 = y0
    elif 3 * pi / 8 < ang <= 5 * pi / 8:  # down direction
        x1 = x0
    elif -5 * pi / 8 < ang <= -3 * pi / 8:  # up direction
        x1 = x0
    elif ang <= -7 * pi / 8 or 7 * pi / 8 < ang:  # left direction
        y1 = y0
    elif pi / 8 < ang <= 3 * pi / 8:  # down-right direction
        if abs(x1 - x0) > abs(y1 - y0):
            x1 = x0 + abs(y1 - y0)
        else:
            y1 = y0 + abs(x1 - x0)
    elif -3 * pi / 8 < ang <= -pi / 8:  # up-left direction
        if abs(x1 - x0) > abs(y1 - y0):
            x1 = x0 + abs(y1 - y0)
        else:
            y1 = y0 - abs(x1 - x0)
    elif 5 * pi / 8 < ang <= 7 * pi / 8:  # down-left direction
        if abs(x1 - x0) > abs(y1 - y0):
            x1 = x0 - abs(y1 - y0)
        else:
            y1 = y0 + abs(x1 - x0)
    elif -7 * pi / 8 < ang <= -5 * pi / 8:  # up-right direction
        if abs(x1 - x0) > abs(y1 - y0):
            x1 = x0 - abs(y1 - y0)
        else:
            y1 = y0 - abs(x1 - x0)
    return QtCore.QPointF(x1, y1)


def _find_nice_rect_position(
    pos: QtCore.QPointF, pos0: QtCore.QPointF
) -> QtCore.QPointF:
    x0, y0 = pos0.x(), pos0.y()
    x1, y1 = pos.x(), pos.y()
    dy, dx = y1 - y0, x1 - x0
    length = min(abs(dy), abs(dx))
    dy_sign = 1 if dy > 0 else -1
    dx_sign = 1 if dx > 0 else -1
    return QtCore.QPointF(x0 + dx_sign * length, y0 + dy_sign * length)



from himena.plugins.widget_class import (
    register_widget_class,
    get_widget_class,
    register_previewer_class,
    widget_classes,
)
from himena.plugins._checker import validate_protocol
from himena.plugins._signature import configure_gui
from himena.plugins.io import (
    register_reader_plugin,
    register_writer_plugin,
    ReaderPlugin,
    WriterPlugin,
)
from himena.plugins.actions import (
    register_function,
    register_hidden_function,
    configure_submenu,
    register_conversion_rule,
    AppActionRegistry,
    ReproduceArgs,
    register_modification_tracker,
    add_default_status_tip,
)
from himena.plugins.widget_plugins import (
    register_dock_widget_action,
    update_config_context,
)
from himena.plugins.install import install_plugins, override_keybindings
from himena.plugins.config import config_field

__all__ = [
    "add_default_status_tip",
    "get_widget_class",
    "register_previewer_class",
    "widget_classes",
    "validate_protocol",
    "configure_gui",
    "get_plugin_interface",
    "install_plugins",
    "config_field",
    "override_keybindings",
    "update_config_context",
    "register_reader_plugin",
    "register_writer_plugin",
    "register_function",
    "register_hidden_function",
    "register_dock_widget_action",
    "register_widget_class",
    "register_conversion_rule",
    "register_modification_tracker",
    "configure_submenu",
    "AppActionRegistry",
    "ReproduceArgs",
    "ReaderPlugin",
    "WriterPlugin",
]



from __future__ import annotations

import re
from typing import (
    TypeVar,
    Iterator,
    NamedTuple,
    Callable,
    overload,
    TYPE_CHECKING,
    Literal,
)
import numpy as np

_C = TypeVar("_C", bound=type)

if TYPE_CHECKING:
    _F = TypeVar("_F", bound=Callable)

    @overload
    def lru_cache(maxsize: int = 128, typed: bool = False) -> Callable[[_F], _F]: ...
    @overload
    def lru_cache(f: _F) -> _F: ...
else:
    from functools import lru_cache  # noqa: F401


def iter_subclasses(cls: _C) -> Iterator[_C]:
    """Recursively iterate over all subclasses of a class."""
    for sub in cls.__subclasses__():
        yield sub
        yield from iter_subclasses(sub)


def is_structured(arr: np.ndarray) -> bool:
    """True if the array is structured."""
    return isinstance(arr.dtype, (np.void, np.dtypes.VoidDType))


_ANSI_BASIC = {
    1: {"font_weight": "bold"},
    2: {"font_weight": "lighter"},
    3: {"font_weight": "italic"},
    4: {"text_decoration": "underline"},
    5: {"text_decoration": "blink"},
    6: {"text_decoration": "blink"},
    8: {"visibility": "hidden"},
    9: {"text_decoration": "line-through"},
}

ANSI_STYLES_LIGHT = {
    **_ANSI_BASIC,
    30: {"color": "white"},
    31: {"color": "red"},
    32: {"color": "green"},
    33: {"color": "teal"},
    34: {"color": "blue"},
    35: {"color": "magenta"},
    36: {"color": "darkblue"},
    37: {"color": "black"},
}

ANSI_STYLES_DARK = {
    **_ANSI_BASIC,
    30: {"color": "black"},
    31: {"color": "orange"},
    32: {"color": "#a0ffae"},
    33: {"color": "yellow"},
    34: {"color": "#8e9dff"},
    35: {"color": "magenta"},
    36: {"color": "cyan"},
    37: {"color": "white"},
}


def ansi2html(
    ansi_string: str,
    is_dark: bool = False,
) -> Iterator[str]:
    """Convert ansi string to colored HTML

    Parameters
    ----------
    ansi_string : str
        text with ANSI color codes.
    styles : dict, optional
        A mapping from ANSI codes to a dict of css kwargs:values,
        by default ANSI_STYLES

    Yields
    ------
    str
        HTML strings that can be joined to form the final html
    """
    previous_end = 0
    styles = ANSI_STYLES_DARK if is_dark else ANSI_STYLES_LIGHT
    in_span = False
    ansi_codes = []
    ansi_finder = re.compile("\033\\[([\\d;]*)([a-zA-Z])")
    for match in ansi_finder.finditer(ansi_string):
        yield ansi_string[previous_end : match.start()]
        previous_end = match.end()
        params, command = match.groups()

        if command not in "mM":
            continue

        try:
            params = [int(p) for p in params.split(";")]
        except ValueError:
            params = [0]

        for i, v in enumerate(params):
            if v == 0:
                params = params[i + 1 :]
                if in_span:
                    in_span = False
                    yield "</span>"
                ansi_codes = []
                if not params:
                    continue

        ansi_codes.extend(params)
        if in_span:
            yield "</span>"
            in_span = False

        if not ansi_codes:
            continue

        style = [
            "; ".join([f"{k}: {v}" for k, v in styles[k].items()]).strip()
            for k in ansi_codes
            if k in styles
        ]
        yield '<span style="{}">'.format("; ".join(style))

        in_span = True

    yield ansi_string[previous_end:]
    if in_span:
        yield "</span>"
        in_span = False


class PluginInfo(NamedTuple):
    """Tuple that describes a plugin function."""

    module: str
    name: str

    def to_str(self) -> str:
        """Return the string representation of the plugin."""
        return f"{self.module}.{self.name}"

    @classmethod
    def from_str(cls, s: str) -> PluginInfo:
        """Create a PluginInfo from a string."""
        mod_name, func_name = s.rsplit(".", 1)
        return PluginInfo(module=mod_name, name=func_name)


def is_subtype(string: str, supertype: str) -> bool:
    """Check if the type is a subtype of the given type.

    ``` python
    is_subtype_of("text", "text")  # True
    is_subtype_of("text.plain", "text")  # True
    is_subtype_of("text.plain", "text.html")  # False
    ```
    """
    string_parts = string.split(".")
    supertype_parts = supertype.split(".")
    if len(supertype_parts) > len(string_parts):
        return False
    return string_parts[: len(supertype_parts)] == supertype_parts


def table_to_text(
    data: np.ndarray,
    format: Literal["CSV", "TSV", "Markdown", "Latex", "rST", "HTML"] = "CSV",
    end_of_text: Literal["", "\n"] = "\n",
) -> tuple[str, str, str]:
    from tabulate import tabulate

    format = format.lower()
    if format == "markdown":
        s = tabulate(data[1:], headers=data[0], tablefmt="github")
        ext_default = ".md"
        language = "markdown"
    elif format == "latex":
        s = _table_to_latex(data)
        ext_default = ".tex"
        language = "latex"
    elif format == "html":
        s = tabulate(data, tablefmt="html")
        ext_default = ".html"
        language = "html"
    elif format == "rst":
        s = tabulate(data, tablefmt="rst")
        ext_default = ".rst"
        language = "rst"
    elif format == "csv":
        s = _to_csv_like(data, ",")
        ext_default = ".csv"
        language = None
    elif format == "tsv":
        s = _to_csv_like(data, "\t")
        ext_default = ".tsv"
        language = None
    else:
        raise ValueError(f"Unknown format: {format}")
    return s + end_of_text, ext_default, language


def _to_csv_like(data: np.ndarray, sep: str) -> str:
    """Convert a table to CSV-like string."""
    return "\n".join(sep.join(str(r) for r in row) for row in data)


def _table_to_latex(table: np.ndarray) -> str:
    """Convert a table to LaTeX."""
    header = table[0]
    body = table[1:]
    latex = "\\begin{tabular}{" + "c" * len(header) + "}\n"
    latex += " & ".join(header) + " \\\\\n"
    for row in body:
        latex += " & ".join(str(r) for r in row) + " \\\\\n"
    latex += "\\hline\n"
    latex += "\\end{tabular}"
    return latex



from __future__ import annotations

from typing import Callable, TypeVar, overload

from himena.types import WidgetClassTuple
from himena.utils.misc import is_subtype
from himena.qt.registry._widgets import QFallbackWidget


# NOTE: Different applications may use different widgets for the same data type.
_APP_TYPE_TO_QWIDGET: dict[str | None, list[WidgetClassTuple]] = {}

_F = TypeVar("_F", bound=type)


@overload
def register_widget_class(
    type_: str,
    widget_class: _F,
    app: str | None,
    priority: int = 100,
) -> _F: ...


@overload
def register_widget_class(
    type_: str,
    widget_class: None,
    app: str | None,
    priority: int = 100,
) -> Callable[[_F], _F]: ...


def register_widget_class(type_, widget_class=None, app=None, priority=100):
    """Register a frontend Qt widget class for the given model type.

    The `__init__` method of the registered class must not take any argument. The class
    must implement `update_model` method to update the widget state from a
    WidgetDataModel.

    ``` python
    @register_widget("text")
    class MyTextEdit(QtW.QPlainTextEdit):
        def update_model(self, model: WidgetDataModel):
            self.setPlainText(model.value)
    ```
    """

    if app is not None and not isinstance(app, str):
        raise TypeError(f"App name must be a string, got {app!r}")
    if not isinstance(type_, str):
        raise TypeError(f"Type must be a string, got {type_!r}")

    def _inner(wdt_class):
        if app not in _APP_TYPE_TO_QWIDGET:
            _APP_TYPE_TO_QWIDGET[app] = []
        _APP_TYPE_TO_QWIDGET[app].append(WidgetClassTuple(type_, wdt_class, priority))
        return wdt_class

    return _inner if widget_class is None else _inner(widget_class)


def list_widget_class(
    app_name: str,
    type: str,
) -> tuple[list[WidgetClassTuple], type[QFallbackWidget]]:
    """List registered widget classes for the given app and super-type."""
    widget_list = _APP_TYPE_TO_QWIDGET.get(None, [])
    if app_name in _APP_TYPE_TO_QWIDGET:
        widget_list = _APP_TYPE_TO_QWIDGET[app_name] + widget_list
    return [
        item for item in widget_list if is_subtype(type, item.type)
    ], QFallbackWidget



from __future__ import annotations
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import Iterator, TYPE_CHECKING

from qtpy import QtWidgets as QtW
from qtpy import QtGui, QtCore
from superqt import QSearchableComboBox, QToggleSwitch

from himena.consts import StandardType, MonospaceFontFamily
from himena.types import WidgetDataModel
from himena.standards.model_meta import TextMeta
from himena.plugins import validate_protocol, config_field

from himena.utils.collections import OrderedSet
from himena.utils.misc import lru_cache
from himena.qt._qcoloredit import QColorSwatch
from himena_builtins.qt.widgets._text_base import QMainTextEdit, POINT_SIZES, TAB_SIZES
from himena_builtins.qt.widgets._shared import labeled

if TYPE_CHECKING:
    from himena.style import Theme

_POPULAR_LANGUAGES = [
    "Plain Text", "Python", "C++", "C", "Java", "JavaScript", "HTML", "CSS", "SQL",
    "Rust", "Go", "TypeScript", "Shell", "Ruby", "PHP", "Swift", "Kotlin", "Dart", "R",
    "Scala", "Perl", "Lua", "Haskell", "Julia", "MATLAB", "Markdown", "YAML", "JSON",
    "XML", "TOML", "PowerShell", "Batch", "C#", "Objective-C",
]  # fmt: skip


@lru_cache(maxsize=1)
def get_languages() -> OrderedSet[str]:
    """Get all languages supported by pygments."""
    from pygments.lexers import get_all_lexers

    langs: OrderedSet[str] = OrderedSet()
    for lang in _POPULAR_LANGUAGES:
        langs.add(lang)
    for lang, aliases, extensions, _ in get_all_lexers(plugins=False):
        langs.add(lang)
    return langs


def find_language_from_path(path: str) -> str | None:
    """Detect language from file path."""
    from pygments.lexers import get_lexer_for_filename
    from pygments.util import ClassNotFound

    # pygment lexer for svg is not available
    if path.endswith(".svg"):
        return "XML"

    try:
        lexer = get_lexer_for_filename(path)
        return lexer.name
    except ClassNotFound:
        return None


class QTextControl(QtW.QWidget):
    languageChanged = QtCore.Signal(str)
    tabChanged = QtCore.Signal(int)

    def __init__(self, text_edit: QMainTextEdit):
        super().__init__()
        self._text_edit = text_edit

        self._language_combobox = QSearchableComboBox()
        self._language_combobox.addItems(get_languages())
        self._language_combobox.setToolTip("Language of the document")
        self._language_combobox.setMaximumWidth(120)
        self._language_combobox.currentIndexChanged.connect(self._emit_language_changed)

        self._tab_spaces_combobox = QtW.QComboBox()
        self._tab_spaces_combobox.addItems([str(x) for x in TAB_SIZES])
        self._tab_spaces_combobox.setCurrentText("4")
        self._tab_spaces_combobox.setToolTip("Tab size")
        self._tab_spaces_combobox.currentTextChanged.connect(
            lambda x: self.tabChanged.emit(int(x))
        )

        self._line_num = QtW.QLabel()
        self._line_num.setFixedWidth(50)

        self._wordwrap = QToggleSwitch("Word Wrap")
        self._wordwrap.setChecked(False)
        self._wordwrap.toggled.connect(self._wordwrap_changed)
        self._wordwrap.setToolTip("Enable word wrap")

        self._encoding = QtW.QLabel("utf-8")

        layout = QtW.QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)
        layout.addWidget(QtW.QWidget())  # spacer
        layout.addWidget(labeled("Ln:", self._line_num))
        layout.addWidget(self._encoding)
        layout.addWidget(self._wordwrap)
        layout.addWidget(labeled("Spaces:", self._tab_spaces_combobox))
        layout.addWidget(labeled("Language:", self._language_combobox))

        # make font smaller
        font = QtGui.QFont()
        font.setFamily(MonospaceFontFamily)
        font.setPointSize(8)
        for child in self.findChildren(QtW.QWidget):
            child.setFont(font)

    def _wordwrap_changed(self, checked: bool):
        """Enable or disable word wrap."""
        if checked:
            mode = QtGui.QTextOption.WrapMode.WordWrap
        else:
            mode = QtGui.QTextOption.WrapMode.NoWrap
        self._text_edit.setWordWrapMode(mode)

    def _emit_language_changed(self):
        self.languageChanged.emit(self._language_combobox.currentText())

    def _move_cursor_to_line(self, line: int):
        cursor = self._text_edit.textCursor()
        cursor.setPosition(0)
        cursor.movePosition(
            QtGui.QTextCursor.MoveOperation.NextBlock,
            QtGui.QTextCursor.MoveMode.KeepAnchor,
            line - 1,
        )
        self._text_edit.setTextCursor(cursor)


class QTextEdit(QtW.QWidget):
    """Default text editor widget."""

    __himena_widget_id__ = "builtins:QTextEdit"
    __himena_display_name__ = "Built-in Text Editor"

    def __init__(self):
        super().__init__()
        self._main_text_edit = QMainTextEdit(self)
        self._control = QTextControl(self._main_text_edit)

        self._control.languageChanged.connect(self._main_text_edit.syntax_highlight)
        self._control.tabChanged.connect(self._main_text_edit.set_tab_size)
        self._main_text_edit.cursorPositionChanged.connect(self._update_line_numbers)
        layout = QtW.QVBoxLayout(self)
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._main_text_edit)
        self._model_type = StandardType.TEXT
        self._extension_default = ".txt"

    @validate_protocol
    def control_widget(self) -> QTextControl:
        return self._control

    @validate_protocol
    def update_configs(self, configs: TextEditConfigs):
        self._main_text_edit._default_font.setPointSize(configs.default_font_size)
        self._main_text_edit.setFont(self._main_text_edit._default_font)
        self._control._tab_spaces_combobox.setCurrentText(str(configs.default_tab_size))

    @validate_protocol
    def update_model(self, model: WidgetDataModel):
        if not isinstance(value := model.value, str):
            value = str(value)
        self._main_text_edit.setPlainText(value)
        lang = None
        spaces = 4
        encoding = None
        if isinstance(model.metadata, TextMeta):
            lang = model.metadata.language
            spaces = model.metadata.spaces
            encoding = model.metadata.encoding
            if sel := model.metadata.selection:
                cursor = self._main_text_edit.textCursor()
                cursor.setPosition(sel[0])
                cursor.setPosition(sel[1], QtGui.QTextCursor.MoveMode.KeepAnchor)
                self._main_text_edit.setTextCursor(cursor)
        if lang is None:
            if isinstance(src := model.source, Path):
                # set default language
                lang = find_language_from_path(src.name)
        # if language could be inferred, set it
        if lang:
            self._control._language_combobox.setCurrentText(lang)
            self._control._emit_language_changed()
        self._control._tab_spaces_combobox.setCurrentText(str(spaces))
        if encoding:
            self._control._encoding.setText(encoding)
        self._model_type = model.type
        if ext := model.extension_default:
            self._extension_default = ext
        return None

    @validate_protocol
    def to_model(self) -> WidgetDataModel[str]:
        cursor = self._main_text_edit.textCursor()
        font = self._main_text_edit.font()
        return WidgetDataModel(
            value=self._main_text_edit.toPlainText(),
            type=self.model_type(),
            extension_default=self._extension_default,
            metadata=TextMeta(
                language=self._control._language_combobox.currentText(),
                spaces=int(self._control._tab_spaces_combobox.currentText()),
                selection=(cursor.selectionStart(), cursor.selectionEnd()),
                encoding=self._control._encoding.text(),
                font_family=font.family(),
                font_size=font.pointSizeF(),
            ),
        )

    @validate_protocol
    def model_type(self):
        return self._model_type

    @validate_protocol
    def size_hint(self) -> tuple[int, int]:
        return 400, 300

    @validate_protocol
    def is_modified(self) -> bool:
        return self._main_text_edit.is_modified()

    @validate_protocol
    def set_modified(self, value: bool) -> None:
        self._main_text_edit.document().setModified(value)

    @validate_protocol
    def is_editable(self) -> bool:
        return not self._main_text_edit.isReadOnly()

    @validate_protocol
    def set_editable(self, value: bool) -> None:
        self._main_text_edit.setReadOnly(not value)

    @validate_protocol
    def widget_added_callback(self):
        self._main_text_edit.setFont(self._main_text_edit._default_font)

    @validate_protocol
    def theme_changed_callback(self, theme: Theme):
        text_edit = self._main_text_edit
        if theme.is_light_background():
            text_edit._code_theme = "default"
        else:
            text_edit._code_theme = "native"
        text_edit.syntax_highlight(text_edit._language)

    def keyPressEvent(self, a0: QtGui.QKeyEvent | None) -> None:
        if (
            a0.key() == QtCore.Qt.Key.Key_F
            and a0.modifiers() & QtCore.Qt.KeyboardModifier.ControlModifier
        ):
            self._main_text_edit._find_string()
            return None
        return super().keyPressEvent(a0)

    def setFocus(self):
        self._main_text_edit.setFocus()

    def _update_line_numbers(self):
        line_num = self._main_text_edit.textCursor().blockNumber() + 1
        self._control._line_num.setText(str(line_num))


class QRichTextEdit(QtW.QWidget):
    __himena_widget_id__ = "builtins:QRichTextEdit"
    __himena_display_name__ = "Built-in HTML Text Editor"

    def __init__(self):
        super().__init__()
        self._main_text_edit = QtW.QTextEdit(self)
        layout = QtW.QVBoxLayout(self)
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._main_text_edit)
        self._control = QRichTextEditControl(self)

    def initPlainText(self, text: str):
        self._main_text_edit.setHtml(text)

    def setFocus(self):
        self._main_text_edit.setFocus()

    @validate_protocol
    def update_model(self, model: WidgetDataModel[str]):
        self.initPlainText(model.value)
        return None

    @validate_protocol
    def to_model(self) -> WidgetDataModel[str]:
        cursor = self._main_text_edit.textCursor()
        font = self._main_text_edit.font()
        return WidgetDataModel(
            value=self._main_text_edit.toHtml(),
            type=self.model_type(),
            extension_default=".html",
            metadata=TextMeta(
                language="HTML",
                selection=(cursor.selectionStart(), cursor.selectionEnd()),
                font_family=font.family(),
                font_size=font.pointSizeF(),
            ),
        )

    @validate_protocol
    def model_type(self):
        return StandardType.HTML

    @validate_protocol
    def size_hint(self) -> tuple[int, int]:
        return 400, 300

    @validate_protocol
    def is_modified(self) -> bool:
        return False

    @validate_protocol
    def set_modified(self, value: bool) -> None:
        self._main_text_edit.document().setModified(value)

    @validate_protocol
    def is_editable(self) -> bool:
        return False

    @validate_protocol
    def set_editable(self, value: bool) -> None:
        self._main_text_edit.setReadOnly(not value)
        self._control.setEnabled(value)

    @validate_protocol
    def control_widget(self) -> QRichTextEditControl:
        return self._control


class QRichTextPushButton(QtW.QPushButton):
    def __init__(self, text: str):
        super().__init__()
        self._label = QtW.QLabel(text, self)
        layout = QtW.QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._label)
        self._label.setAttribute(QtCore.Qt.WidgetAttribute.WA_TransparentForMouseEvents)
        self._label.setAttribute(QtCore.Qt.WidgetAttribute.WA_TranslucentBackground)
        self._label.setSizePolicy(
            QtW.QSizePolicy.Policy.Expanding,
            QtW.QSizePolicy.Policy.Expanding,
        )
        self._label.setAlignment(
            QtCore.Qt.AlignmentFlag.AlignCenter | QtCore.Qt.AlignmentFlag.AlignVCenter
        )


def _make_btn(text: str, tooltip: str, callback) -> QtW.QPushButton:
    btn = QRichTextPushButton(text)
    btn.setFixedWidth(20)
    btn.setToolTip(tooltip)
    btn.clicked.connect(callback)
    return btn


def _make_color_swatch(label: str, tooltip: str, callback) -> QtW.QWidget:
    group = QtW.QWidget()
    layout = QtW.QHBoxLayout(group)
    layout.setContentsMargins(0, 0, 0, 0)
    swatch = QColorSwatch()
    swatch.setQColor(QtGui.QColor(0, 0, 0, 0))
    swatch.setToolTip(tooltip)
    swatch.setFixedSize(25, 20)
    swatch.colorChanged.connect(callback)
    label_widget = QtW.QLabel(label)
    label_widget.setAlignment(
        QtCore.Qt.AlignmentFlag.AlignRight | QtCore.Qt.AlignmentFlag.AlignVCenter
    )
    label_widget.setFixedWidth(20)
    layout.addWidget(label_widget)
    layout.addWidget(swatch)
    return group


class QRichTextEditControl(QtW.QWidget):
    def __init__(self, text_edit: QRichTextEdit):
        super().__init__()
        self._foreground_color_edit = _make_color_swatch(
            label="FG:",
            tooltip="Change Foreground Color",
            callback=self._on_foreground_color_changed,
        )
        self._background_color_edit = _make_color_swatch(
            label="BG:",
            tooltip="Change Background Color",
            callback=self._on_background_color_changed,
        )
        self._toggle_bold_button = _make_btn(
            "<b>B</b>",
            tooltip="Toggle Bold",
            callback=self._on_toggle_bold,
        )
        self._toggle_it_button = _make_btn(
            "<i>I</i>",
            tooltip="Toggle Italic",
            callback=self._on_toggle_italic,
        )
        self._toggle_underline_button = _make_btn(
            "<u>U</u>",
            tooltip="Toggle Underline",
            callback=self._on_toggle_underline,
        )
        self._toggle_strike_button = _make_btn(
            "<s>S</s>", tooltip="Toggle Strike", callback=self._on_toggle_strike
        )
        spacer = QtW.QWidget()
        spacer.setSizePolicy(
            QtW.QSizePolicy.Policy.Expanding, QtW.QSizePolicy.Policy.Preferred
        )
        layout = QtW.QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(spacer)
        layout.addWidget(self._foreground_color_edit)
        layout.addWidget(self._background_color_edit)
        layout.addWidget(self._toggle_bold_button)
        layout.addWidget(self._toggle_it_button)
        layout.addWidget(self._toggle_underline_button)
        layout.addWidget(self._toggle_strike_button)

        self._text_edit = text_edit

    def _on_foreground_color_changed(self, color: QtGui.QColor):
        with self._merging_format() as fmt:
            fmt.setForeground(color)

    def _on_background_color_changed(self, color: QtGui.QColor):
        with self._merging_format() as fmt:
            fmt.setBackground(color)

    def _on_toggle_bold(self):
        with self._merging_format() as fmt:
            fmt.setFontWeight(
                QtGui.QFont.Weight.Bold
                if not fmt.font().bold()
                else QtGui.QFont.Weight.Normal
            )

    def _on_toggle_italic(self):
        with self._merging_format() as fmt:
            fmt.setFontItalic(not fmt.font().italic())

    def _on_toggle_underline(self):
        with self._merging_format() as fmt:
            fmt.setFontUnderline(not fmt.font().underline())

    def _on_toggle_strike(self):
        with self._merging_format() as fmt:
            fmt.setFontStrikeOut(not fmt.font().strikeOut())

    @contextmanager
    def _merging_format(self) -> Iterator[QtGui.QTextCharFormat]:
        cursor = self._text_edit._main_text_edit.textCursor()
        fmt = cursor.charFormat()
        yield fmt
        cursor.mergeCharFormat(fmt)
        self._text_edit._main_text_edit.setTextCursor(cursor)


@dataclass
class TextEditConfigs:
    default_font_size: int = config_field(
        default=10,
        label="Default font size (pt).",
        choices=POINT_SIZES,
    )
    default_tab_size: int = config_field(
        default=4,
        label="Default tab size.",
        choices=TAB_SIZES,
    )



from __future__ import annotations

from typing import TYPE_CHECKING, Iterable, Iterator
from logging import getLogger
from contextlib import contextmanager, suppress
from concurrent.futures import Future, ThreadPoolExecutor
import warnings
from qtpy import QtWidgets as QtW, QtCore, QtGui
from qtpy.QtCore import Qt
from app_model.types import KeyBindingRule
from app_model.backends.qt import QCommandAction, QModelMenu, QKeyBindingSequence
from superqt import ensure_main_thread

from himena.consts import MonospaceFontFamily

if TYPE_CHECKING:
    from himena._app_model import HimenaApplication
    from himena.qt.main_window import MainWindowQt

_EXECUTOR = ThreadPoolExecutor()
_LOGGER = getLogger(__name__)


class H:
    TITLE = 0
    KEYBINDING = 1
    WHEN = 2
    SOURCE = 3
    COMMAND_ID = 4
    WEIGHT = 5


class QKeybindEdit(QtW.QWidget):
    """Widget for editing keybindings."""

    def __init__(self, ui: MainWindowQt):
        super().__init__()
        self._ui = ui
        layout = QtW.QVBoxLayout(self)
        self._search = QKeybindSearch(self)
        layout.addWidget(self._search)
        self._table = QKeybindTable(ui.model_app, self)
        layout.addWidget(self._table)
        self._table.update_table_from_model_app(ui.model_app)
        self._restore_default_btn = QtW.QPushButton("Restore")
        self._restore_default_btn.setToolTip("Restore the default keybindings.")
        self._restore_default_btn.setFixedWidth(84)
        self._restore_default_btn.clicked.connect(self._restore_default)
        layout.addWidget(
            self._restore_default_btn, alignment=Qt.AlignmentFlag.AlignRight
        )
        self._search.textChanged.connect(self._on_search_text_changed)
        self._table.keybinding_updated.connect(self._on_keybinding_updated)

    def _on_search_text_changed(self, text: str):
        self._table.filter_by_text_async(text)

    def _on_keybinding_updated(self, command_id: str, new_shortcut: str):
        _LOGGER.info("Keybindings registered.")
        self._table.update_table_from_model_app(self._ui.model_app)
        self._table.filter_by_text(self._search.text())  # re-filter
        self._ui.app_profile.with_keybinding_override(new_shortcut, command_id).save()
        qui = self._ui._backend_main_window
        for qa in _iter_command_action(qui._menubar.actions() + qui._toolbar.actions()):
            if qa._command_id == command_id:
                qa.setShortcut(new_shortcut)

    def _restore_default(self):
        _LOGGER.info("Restoring keybindings.")
        qui = self._ui._backend_main_window
        app = self._ui.model_app
        for ko in self._ui.app_profile.keybinding_overrides:
            keybind = app.keybindings.get_keybinding(ko.command_id)
            if keybind:
                app.keybindings._keymap.pop(keybind.keybinding.to_int(), -1)
        self._ui.app_profile.keybinding_overrides.clear()
        self._ui.app_profile.save()
        for qa in _iter_command_action(qui._menubar.actions() + qui._toolbar.actions()):
            if action := self._ui.model_app.registered_actions.get(qa._command_id):
                if action.keybindings:
                    for kb in action.keybindings:
                        if kb.primary:
                            app.keybindings.register_keybinding_rule(
                                action.id, KeyBindingRule(primary=kb.primary)
                            )
                            kb_obj = app.keybindings.get_keybinding(
                                action.id
                            ).keybinding
                            qa.setShortcut(QKeyBindingSequence(kb_obj))
        self._table.update_table_from_model_app(self._ui.model_app)
        self._table.filter_by_text(self._search.text())  # re-filter


class QKeybindSearch(QtW.QLineEdit):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setPlaceholderText("Type to Search keybindings...")
        self.setToolTip(
            "Search keybindings by command title, keybinding, or command ID."
        )


class QKeybindTable(QtW.QTableWidget):
    keybinding_updated = QtCore.Signal(str, str)  # command_id, new_keybinding

    def __init__(self, app: HimenaApplication, parent=None):
        super().__init__(parent)
        self._app = app
        self.setColumnCount(5)
        self.setSelectionBehavior(QtW.QAbstractItemView.SelectionBehavior.SelectRows)
        self.setSelectionMode(QtW.QAbstractItemView.SelectionMode.SingleSelection)

        # Design and UX
        self.setShowGrid(False)
        self.horizontalHeader().setStretchLastSection(True)
        self.setHorizontalHeaderLabels(
            ["Command", "Keybinding", "When", "Source", "ID", "Weight"]
        )
        self.verticalHeader().setVisible(False)
        self._default_row_height = 22
        self.horizontalHeader().setFixedHeight(self._default_row_height)
        self.setVerticalScrollMode(QtW.QAbstractItemView.ScrollMode.ScrollPerPixel)
        self.setHorizontalScrollMode(QtW.QAbstractItemView.ScrollMode.ScrollPerPixel)

        # Hide "weight" for now, but may be useful in the future
        self.setColumnHidden(H.WEIGHT, True)

        self.setItemDelegateForColumn(H.KEYBINDING, QKeybindDelegate(self))
        self.setColumnWidth(H.TITLE, 210)
        self.setColumnWidth(H.KEYBINDING, 120)
        self.setColumnWidth(H.WHEN, 120)
        self.setColumnWidth(H.WEIGHT, 40)
        self.setColumnWidth(H.SOURCE, 60)
        self.setColumnWidth(H.COMMAND_ID, 180)

        self._update_blocked = False
        self._last_future: Future | None = None
        self.cellChanged.connect(self._update_keybinding)

    def update_table_from_model_app(self, app: HimenaApplication):
        commands_to_skip = app._dynamic_command_ids
        commands = sorted(
            (cmd[1] for cmd in app.commands if cmd[1].id not in commands_to_skip),
            key=lambda cmd: cmd.title,
        )
        nrows = len(commands)
        self.setRowCount(nrows)
        with self.block_context():
            self.clearContents()
            self.setRowCount(len(commands))
            for i, cmd in enumerate(commands):
                self.setRowHeight(i, self._default_row_height)
                cmd_id = cmd.id
                if ":" in cmd_id:
                    source = cmd_id.split(":", maxsplit=1)[0]
                else:
                    source = ""
                title = cmd.title
                kbd = app.keybindings.get_keybinding(cmd_id)
                if kbd:
                    key_seq = kbd.keybinding.to_text()
                    when = str(kbd.when) if kbd.when is not None else ""
                    weight = str(kbd.weight)
                else:
                    key_seq = ""
                    when = ""
                    weight = ""
                self.setItem(i, H.TITLE, _item_basic(title))
                self.setItem(i, H.KEYBINDING, QtW.QTableWidgetItem(key_seq))
                self.setItem(i, H.WHEN, _item_basic(when, monospace=True))
                self.setItem(i, H.WEIGHT, _item_basic(weight))
                self.setItem(i, H.SOURCE, _item_basic(source))
                self.setItem(i, H.COMMAND_ID, _item_basic(cmd_id, monospace=True))
        return None

    def filter_by_text(self, text: str) -> list[bool] | None:
        text = text.lower().strip()

        # disable filtering if text is empty
        if text == "":
            return None

        parts = text.split(" ")
        out = []
        for row in range(self.rowCount()):
            kb = self.item(row, H.KEYBINDING).text().lower()
            title = self.item(row, H.TITLE).text().lower()
            id = self.item(row, H.COMMAND_ID).text().lower()
            if (
                kb
                and any(part in kb for part in parts)
                or title
                and any(part in title for part in parts)
                or id
                and any(part in id for part in parts)
            ):
                out.append(True)
            else:
                out.append(False)
        return out

    def filter_by_text_async(self, text: str) -> Future:
        if self._last_future:
            self._last_future.cancel()
        future = _EXECUTOR.submit(self.filter_by_text, text)
        self._last_future = future
        future.add_done_callback(self._filter_done)
        return future

    @ensure_main_thread
    def _filter_done(self, future: Future):
        if future.cancelled():
            return
        result = future.result()
        self._last_future = None
        with suppress(RuntimeError):
            # this function may be called after the widget is destroyed
            if result is not None:
                for i, show in enumerate(result):
                    self.setRowHidden(i, not show)
            else:
                for i in range(self.rowCount()):
                    self.setRowHidden(i, False)
        return None

    @contextmanager
    def block_context(self):
        was_blocked = self._update_blocked
        self._update_blocked = True
        try:
            yield
        finally:
            self._update_blocked = was_blocked

    def _update_keybinding(self, row: int, col: int):
        if col != H.KEYBINDING or self._update_blocked:
            return
        self.setCurrentItem(self.item(row, col))
        current_item = self.currentItem()
        if current_item is None:
            return
        new_shortcut = current_item.text()
        command_id = self.item(row, H.COMMAND_ID).text()
        kbd_current = self._app.keybindings.get_keybinding(command_id)
        if kbd_current is None or kbd_current.keybinding.to_text() != new_shortcut:
            conflictions = self._get_confliction_command_ids(new_shortcut, command_id)
            if conflictions:
                warnings.warn(
                    f"Keybinding {new_shortcut} probably conflicts with the following "
                    f"commands: {conflictions}",
                    RuntimeWarning,
                    stacklevel=2,
                )
            if kbd_current is not None:
                self._app.keybindings._keymap.pop(kbd_current.keybinding.to_int(), -1)
            if new_shortcut:
                self._app.keybindings.register_keybinding_rule(
                    command_id, KeyBindingRule(primary=new_shortcut)
                )
            self.keybinding_updated.emit(command_id, new_shortcut)

    def _get_confliction_command_ids(
        self, keybinding: str, except_for: str
    ) -> list[str]:
        conflictions: list[str] = []
        for kbd in self._app.keybindings:
            if kbd.command_id == except_for:
                continue
            if kbd.keybinding.to_text() == keybinding:
                # TODO: check "when" to avoid adding conflictions of independent
                # keybindings.
                conflictions.append(kbd.command_id)
        return conflictions


def _item_basic(text: str, monospace: bool = False) -> QtW.QTableWidgetItem:
    item = QtW.QTableWidgetItem(text)
    item.setToolTip(text)
    item.setFlags(item.flags() & ~Qt.ItemFlag.ItemIsEditable)
    if monospace:
        item.setFont(QtGui.QFont(MonospaceFontFamily))
    return item


class QKeybindDelegate(QtW.QItemDelegate):
    """Delegate that handles when user types in new shortcut."""

    def createEditor(self, widget, style_option, model_index: QtCore.QModelIndex):
        if model_index.isValid():
            self._editor = QKeybindingLineEdit(widget)
            return self._editor

    def setEditorData(
        self, widget: QKeybindingLineEdit, model_index: QtCore.QModelIndex
    ):
        if model_index.isValid():
            text = widget.keySequence().toString()
            widget.setKeySequence(text)

    def updateEditorGeometry(
        self,
        widget: QKeybindingLineEdit,
        style_option: QtW.QStyleOptionViewItem,
        model_index: QtCore.QModelIndex,
    ):
        if model_index.isValid():
            widget.setGeometry(style_option.rect)

    def setModelData(
        self,
        widget: QKeybindingLineEdit,
        abstract_item_model: QtCore.QAbstractItemModel,
        model_index: QtCore.QModelIndex,
    ):
        if model_index.isValid():
            text = widget.keySequence().toString()
            abstract_item_model.setData(model_index, text, Qt.ItemDataRole.EditRole)


class QKeybindingLineEdit(QtW.QKeySequenceEdit):
    """Widget used to edit keybindings."""


def _find_action(
    actions: Iterable[QtW.QAction], command_id: str
) -> Iterator[QCommandAction]:
    """Yield all actions with the given command ID."""
    for action in actions:
        if isinstance(action, QCommandAction):
            if action._command_id == command_id:
                yield action
            elif menu := action.menu():
                yield from _find_action(menu.actions(), command_id)
        elif isinstance(menu := action.menu(), QModelMenu):
            yield from _find_action(menu.actions(), command_id)


def _iter_command_action(actions: Iterable[QtW.QAction]) -> Iterator[QCommandAction]:
    for action in actions:
        if isinstance(action, QCommandAction):
            yield action
        elif isinstance(menu := action.menu(), QModelMenu):
            yield from _iter_command_action(menu.actions())



from __future__ import annotations

from typing import TYPE_CHECKING
from himena.standards import plotting as hplt
from himena_builtins.qt.plot._register import convert_plot_model

if TYPE_CHECKING:
    from matplotlib import pyplot as plt
    from mpl_toolkits import mplot3d as plt3d


def _refer_x_axis(ax: hplt.Axis, ax_mpl: plt.Axes):
    if ax.label is not None:
        ax_mpl.set_xlabel(ax.label)
    if ax.ticks is not None:
        ax_mpl.set_xticks(ax.ticks)
    if ax.lim is not None:
        ax_mpl.set_xlim(ax.lim)
    if ax.scale == "log":
        ax_mpl.set_xscale("log")


def _refer_y_axis(ax: hplt.Axis, ax_mpl: plt.Axes):
    if ax.label is not None:
        ax_mpl.set_ylabel(ax.label)
    if ax.ticks is not None:
        ax_mpl.set_yticks(ax.ticks)
    if ax.lim is not None:
        ax_mpl.set_ylim(ax.lim)
    if ax.scale == "log":
        ax_mpl.set_yscale("log")


def _refer_z_axis(ax: hplt.Axis, ax_mpl: plt3d.Axes3D):
    if ax.label is not None:
        ax_mpl.set_zlabel(ax.label)
    if ax.ticks is not None:
        ax_mpl.set_zticks(ax.ticks)
    if ax.lim is not None:
        ax_mpl.set_zlim(ax.lim)
    if ax.scale == "log":
        ax_mpl.set_zscale("log")


def _refer_title(ax: hplt.Axes, ax_mpl: plt.Axes):
    title, style = _parse_styled_text(ax.title)
    ax_mpl.set_title(title, **style)


def update_mpl_axes_by_model(ax: hplt.Axes, ax_mpl: plt.Axes):
    ax_mpl.spines["top"].set_visible(False)
    ax_mpl.spines["right"].set_visible(False)
    ax_mpl.spines["left"].set_edgecolor(ax.axis_color)
    ax_mpl.spines["bottom"].set_edgecolor(ax.axis_color)
    if ax.title is not None:
        _refer_title(ax, ax_mpl)
    if ax.x is not None:
        _refer_x_axis(ax.x, ax_mpl)
    if ax.y is not None:
        _refer_y_axis(ax.y, ax_mpl)
    for model in ax.models:
        convert_plot_model(model, ax_mpl)


def _convert_axes_3d(ax: hplt.Axes3D, ax_mpl: plt3d.Axes3D):
    if ax.title is not None:
        _refer_title(ax, ax_mpl)
    if ax.x is not None:
        _refer_x_axis(ax.x, ax_mpl)
    if ax.y is not None:
        _refer_y_axis(ax.y, ax_mpl)
    if ax.z is not None:
        _refer_z_axis(ax.z, ax_mpl)
    for model in ax.models:
        convert_plot_model(model, ax_mpl)


def _fill_axis_props(axes: hplt.Axes, axes_mpl: plt.Axes):
    """Fill Nones in model axis from the matplotlib axis."""
    axes = axes.model_copy()
    if axes.title is None:
        axes.title = axes_mpl.get_title()
    if axes.x.lim is None:
        axes.x.lim = axes_mpl.get_xlim()
    if axes.x.label is None:
        axes.x.label = axes_mpl.get_xlabel()
    if axes.y.lim is None:
        axes.y.lim = axes_mpl.get_ylim()
    if axes.y.label is None:
        axes.y.label = axes_mpl.get_ylabel()
    return axes


def update_model_axis_by_mpl(axes: hplt.Axes, axes_mpl: plt.Axes):
    """Update model axis from the matplotlib axis."""
    axes.title = axes_mpl.get_title()
    axes.x.lim = axes_mpl.get_xlim()
    axes.x.label = axes_mpl.get_xlabel()
    axes.y.lim = axes_mpl.get_ylim()
    axes.y.label = axes_mpl.get_ylabel()
    return axes


def convert_plot_layout(lo: hplt.BaseLayoutModel, fig: plt.Figure):
    fig.patch.set_facecolor(lo.background_color)
    if isinstance(lo, hplt.SingleAxes):
        ax_mpl = _get_single_mpl_axes(fig)
        update_mpl_axes_by_model(lo.axes, ax_mpl)
        ax_mpl.patch.set_color(lo.background_color)
        lo.axes = _fill_axis_props(lo.axes, ax_mpl)
    elif isinstance(lo, hplt.layout.Layout1D):
        _shape = (1, len(lo.axes)) if isinstance(lo, hplt.Row) else (len(lo.axes), 1)
        if len(fig.axes) != len(lo.axes):
            fig.clear()
            axes = fig.subplots(*_shape, sharex=lo.share_x, sharey=lo.share_y)
        else:
            axes = fig.axes
        for ax, ax_mpl in zip(lo.axes, axes):
            update_mpl_axes_by_model(ax, ax_mpl)
            ax_mpl.patch.set_color(lo.background_color)
        lo.axes = [_fill_axis_props(ax, ax_mpl) for ax, ax_mpl in zip(lo.axes, axes)]
    elif isinstance(lo, hplt.Grid):
        raise NotImplementedError("Grid layout is not supported yet")
    elif isinstance(lo, hplt.SingleAxes3D):
        ax_mpl = _get_single_mpl_axes(fig, projection="3d")
        _convert_axes_3d(lo.axes, ax_mpl)
        ax_mpl.patch.set_color(lo.background_color)
        lo.axes = _fill_axis_props(lo.axes, ax_mpl)
    elif isinstance(lo, hplt.SingleStackedAxes):
        ax_mpl = _get_single_mpl_axes(fig)
        ax_mpl.patch.set_color(lo.background_color)
    else:
        raise ValueError(f"Unsupported layout model: {lo}")
    return lo


def _get_single_mpl_axes(fig: plt.Figure, **kwargs) -> plt.Axes:
    if len(fig.axes) != 1:
        fig.clear()
        ax_mpl = fig.add_subplot(111, **kwargs)
    else:
        ax_mpl = fig.axes[0]
        ax_mpl.clear()
    return ax_mpl


def _parse_styled_text(text: hplt.StyledText | str) -> tuple[str, dict]:
    if isinstance(text, str):
        return text, {}
    fontdict = {}
    if text.size is not None:
        fontdict["size"] = text.size
    if text.family:
        fontdict["family"] = text.family
    if text.bold:
        fontdict["weight"] = "bold"
    if text.italic:
        fontdict["style"] = "italic"
    if text.underline:
        fontdict["decoration"] = "underline"
    if text.color:
        fontdict["color"] = text.color
    loc = text.alignment
    return text.text, {"fontdict": fontdict, "loc": loc}



import ast
import csv
import inspect
from io import StringIO
from types import FunctionType
from himena.plugins import register_function, configure_gui, configure_submenu
from himena.types import Parametric, WidgetDataModel
from himena.standards.model_meta import TextMeta, FunctionMeta
from himena.consts import StandardType, MenuId
from himena import create_model

configure_submenu(MenuId.TOOLS_TEXT, group="20_builtins", order=0)


@register_function(
    types=StandardType.TEXT,
    menus=[MenuId.TOOLS_TEXT],
    keybindings="Ctrl+F5",
    command_id="builtins:text:run-script",
    group="text-run",
)
def run_script(model: WidgetDataModel[str]):
    """Run a Python script."""
    script = model.value
    if _maybe_python(model):
        exec(script)
    else:
        raise ValueError("Cannot run non-Python script.")
    return None


@register_function(
    title="Run Main Function ...",
    menus=[MenuId.TOOLS_TEXT],
    types=StandardType.TEXT,
    command_id="builtins:text:run-script-main",
    group="text-run",
)
def run_main_function(model: WidgetDataModel[str]) -> Parametric:
    """Run the main function of a Python script with parameters."""
    script = model.value
    if not _maybe_python(model):
        raise ValueError("Cannot run non-Python script.")
    ns = {"create_model": create_model}
    loc = {}
    exec(script, ns, loc)
    if not callable(main := loc.get("main")):
        raise ValueError("No callable `main` function found in the script.")
    gui_options = {}
    for param in inspect.signature(main).parameters.values():
        if param.annotation in (
            inspect.Parameter.empty,
            WidgetDataModel,
            "WidgetDataModel",
        ):
            annot = WidgetDataModel
        else:
            annot = param.annotation
        gui_options[param.name] = {"annotation": annot, "label": param.name}
        if param.default is not inspect.Parameter.empty:
            gui_options[param.name]["default"] = param.default

    return configure_gui(main, gui_options=gui_options)


@register_function(
    types=StandardType.TEXT,
    menus=[MenuId.TOOLS_TEXT],
    command_id="builtins:text:change-separator",
)
def change_separator(model: WidgetDataModel[str]) -> Parametric:
    """Change the separator (in the sense of CSV or TSV) of a text."""

    def change_separator_data(
        old: str = ",", new: str = r"\t", inplace: bool = False
    ) -> WidgetDataModel[str]:
        if old == "" or new == "":
            raise ValueError("Old and new separators must not be empty.")
        # decode unicode escape. e.g., "\\t" -> "\t"
        old = old.encode().decode("unicode_escape")
        new = new.encode().decode("unicode_escape")
        buf = StringIO(model.value)
        reader = csv.reader(buf, delimiter=old)
        new_text = "\n".join(new.join(row) for row in reader)
        return WidgetDataModel(
            value=new_text,
            type=model.type,
            extensions=model.extensions,
            metadata=model.metadata,
            update_inplace=inplace,
        ).with_title_numbering()

    return change_separator_data


@register_function(
    types=StandardType.TEXT,
    menus=[MenuId.TOOLS_TEXT],
    command_id="builtins:text:change-encoding",
)
def change_encoding(model: WidgetDataModel[str]) -> Parametric:
    """Change the encoding of a text."""

    def change_encoding_data(encoding: str = "utf-8") -> WidgetDataModel[str]:
        new_text = model.value.encode(encoding).decode(encoding)
        out = model.with_value(new_text, update_inplace=True)
        if isinstance(meta := model.metadata, TextMeta):
            meta.encoding = encoding
        return out

    return change_encoding_data


@register_function(
    title="Compile As a Function",
    types=StandardType.TEXT,
    menus=[MenuId.TOOLS_TEXT],
    command_id="builtins:text:compile-as-function",
    group="text-run",
)
def compile_as_function(model: WidgetDataModel[str]) -> WidgetDataModel:
    """Compile a Python script as a function data."""
    code = model.value
    if not _maybe_python(model):
        raise ValueError("Cannot compile non-Python script.")
    mod = ast.parse(code)
    global_vars = {}
    local_vars = {}
    out = None
    nblock = len(mod.body)
    filename = "<QFunctionEdit>"
    if nblock == 0:
        raise ValueError("Code is empty.")
    last = mod.body[-1]
    if isinstance(last, ast.Expr):
        if len(mod.body) > 1:
            block_pre = ast.Module(body=mod.body[:-1], type_ignores=[])
            exec(compile(block_pre, filename, "exec"), global_vars, local_vars)
        if isinstance(last.value, ast.Name):
            out = local_vars[last.value.id]
        else:
            out = eval(compile(last.value, filename, "eval"), global_vars, local_vars)
    elif isinstance(last, ast.FunctionDef):
        exec(compile(mod, filename, "exec"), global_vars, local_vars)
        out = local_vars[last.name]
        assert isinstance(out, FunctionType)
        out.__globals__.update(local_vars)
    else:
        raise ValueError("Code must ends with an expression or a function definition.")

    if not callable(out):
        raise ValueError("Code does not define a callable object.")

    return WidgetDataModel(
        value=out,
        type=StandardType.FUNCTION,
        metadata=FunctionMeta(source_code=code),
    )


def _maybe_python(model: WidgetDataModel[str]) -> bool:
    if isinstance(model.metadata, TextMeta):
        lang = model.metadata.language or ""
        return lang.lower() in ("python", "plain text", "")
    return True



from __future__ import annotations

from qtpy import QtWidgets as QtW, QtGui
from qtpy.QtCore import Qt
from typing import TYPE_CHECKING, Generic, TypeVar
import itertools

_W = TypeVar("_W", bound=QtW.QPlainTextEdit)
_X = TypeVar("_W", bound=QtW.QWidget)


class _QFinderBaseWidget(QtW.QDialog, Generic[_X]):
    def __init__(self, parent: _W):
        super().__init__(parent, Qt.WindowType.SubWindow)
        _layout = QtW.QHBoxLayout(self)
        _layout.setContentsMargins(2, 2, 2, 2)
        self.setLayout(_layout)
        _line = QtW.QLineEdit()
        _btn_prev = QtW.QPushButton("▲")
        _btn_next = QtW.QPushButton("▼")
        _btn_prev.setFixedSize(18, 18)
        _btn_next.setFixedSize(18, 18)
        _layout.addWidget(_line)
        _layout.addWidget(_btn_prev)
        _layout.addWidget(_btn_next)
        _btn_prev.clicked.connect(self._btn_prev_clicked)
        _btn_next.clicked.connect(self._btn_next_clicked)
        _line.textChanged.connect(self._find_update)
        self._line_edit = _line
        self._btn_prev = _btn_prev
        self._btn_next = _btn_next
        self._parent_widget = parent

    # fmt: off
    if TYPE_CHECKING:
        def parentWidget(self) -> _W: ...
    # fmt: on

    def show(self):
        super().show()
        self._line_edit.setFocus()

    def _btn_prev_clicked(self):
        self._find_prev()
        self._line_edit.setFocus()

    def _btn_next_clicked(self):
        self._find_next()
        self._line_edit.setFocus()

    def _find_prev(self):
        raise NotImplementedError

    def _find_next(self):
        raise NotImplementedError

    def _find_update(self):
        raise NotImplementedError

    def keyPressEvent(self, a0: QtGui.QKeyEvent) -> None:
        if a0.key() == Qt.Key.Key_Escape:
            self.hide()
            self.parentWidget().setFocus()
        elif a0.key() in (Qt.Key.Key_Enter, Qt.Key.Key_Return):
            if a0.modifiers() & Qt.KeyboardModifier.ShiftModifier:
                self._find_prev()
            else:
                self._find_next()
        return super().keyPressEvent(a0)


class QFinderWidget(_QFinderBaseWidget[_W]):
    """A finder widget for a text editor."""

    def _find_prev(self):
        text = self._line_edit.text()
        if text == "":
            return
        qtext = self.parentWidget()
        flag = QtGui.QTextDocument.FindFlag.FindBackward
        found = qtext.find(text, flag)
        if not found:
            qtext.moveCursor(QtGui.QTextCursor.MoveOperation.End)
            qtext.find(text, flag)

    def _find_next(self):
        text = self._line_edit.text()
        if text == "":
            return
        qtext = self.parentWidget()
        found = qtext.find(text)
        if not found:
            qtext.moveCursor(QtGui.QTextCursor.MoveOperation.Start)
            qtext.find(text)

    _find_update = _find_next


class QTableFinderWidget(_QFinderBaseWidget[QtW.QTableView]):
    def _find_prev(self):
        line_text = self._line_edit.text()
        if line_text == "":
            return
        i, nr, nc = self._get_current_state()
        i -= 1

        for ith in itertools.chain(range(i, -1, -1), range(nr * nc - 1, i, -1)):
            if self._run_until_found(line_text, ith, nc):
                return

    def _find_next(self):
        line_text = self._line_edit.text()
        if line_text == "":
            return
        i, nr, nc = self._get_current_state()
        i += 1

        for ith in itertools.chain(range(i, nr * nc), range(i)):
            if self._run_until_found(line_text, ith, nc):
                return

    def _find_update(self):
        line_text = self._line_edit.text()
        if line_text == "":
            return
        i, nr, nc = self._get_current_state()

        for ith in itertools.chain(range(i, nr * nc), range(i)):
            if self._run_until_found(line_text, ith, nc):
                return

    def _get_current_state(self) -> tuple[int, int, int]:
        qtable = self._parent_widget
        index = qtable.currentIndex()
        model = qtable.model()
        i = index.row() * model.columnCount() + index.column()
        nr, nc = model.rowCount(), model.columnCount()
        return i, nr, nc

    def _run_until_found(self, line_text: str, ith: int, nc: int) -> bool:
        r, c = divmod(ith, nc)
        model = self._parent_widget.model()
        index = model.index(r, c)
        displayed: str = model.data(index, Qt.ItemDataRole.DisplayRole)
        if not isinstance(displayed, str):
            return False

        if displayed == "":
            return False
        if line_text in displayed:
            self._parent_widget.setCurrentIndex(index)
            return True
        return False
